/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

#ifndef DUNE_ASH_GLUE_L2PROJECTION_HH
#define DUNE_ASH_GLUE_L2PROJECTION_HH

#include <dune/geometry/quadraturerules.hh>

#include <dune/fem/quadrature/elementquadrature.hh>
#include <dune/fem/operator/1order/localmassmatrix.hh>

namespace Dune
{

  namespace Ash
  {

    template< class GridPartGlue >
    class GluedL2Projection
    {
      typedef GluedL2Projection< GridPartGlue > ThisType;

    public:
      typedef GridPartGlue GridPartGlueType;

    private:
      typedef typename GridPartGlueType::ctype ctype;

      static const int dimension = GridPartGlueType::dimension;

      typedef typename GridPartGlueType::IntersectionIterator IntersectionIterator;
      typedef typename GridPartGlueType::Intersection Intersection;

      typedef typename GridPartGlueType::InsideGridPartType DomainGridPart;
      typedef typename GridPartGlueType::OutsideGridPartType RangeGridPart;

      typedef typename Intersection::InsideEntityPointer DomainEntityPointer;
      typedef typename Intersection::OutsideEntityPointer RangeEntityPointer;

      typedef Dune::QuadratureRule< ctype, dimension > QuadratureRule;
      typedef Dune::QuadratureRules< ctype, dimension > QuadratureRules;
      typedef typename QuadratureRule::const_iterator QuadratureIterator;

    public:
      explicit GluedL2Projection ( const GridPartGlueType &glue )
      : glue_( glue )
      {}

      template< class DomainFunction, class RangeFunction >
      void operator() ( const DomainFunction &u, RangeFunction &w ) const
      {
        typedef typename DomainFunction::LocalFunctionType DomainLocalFunction;
        typedef typename RangeFunction::LocalFunctionType RangeLocalFunction;

        typedef typename RangeFunction::DiscreteFunctionSpaceType RangeSpace;
        typedef typename RangeSpace::IteratorType RangeIterator;

        typedef Fem::LocalMassMatrix< RangeSpace, Fem::ElementQuadrature< RangeGridPart, 0 > > LocalMassMatrix;

        // assemle right hand side into w

        w.clear();

        const IntersectionIterator iend = glue_.end();
        for( IntersectionIterator iit = glue_.begin(); iit != iend; ++iit )
        {
          const Intersection &intersection = *iit;
          const typename Intersection::Geometry geometry = intersection.geometry();

          const DomainEntityPointer pInside = intersection.inside();
          const typename DomainEntityPointer::Entity &inside = *pInside;
          const DomainLocalFunction ulocal = u.localFunction( inside );

          const RangeEntityPointer pOutside = intersection.outside();
          const typename RangeEntityPointer::Entity &outside= *pOutside;
          RangeLocalFunction wlocal = w.localFunction( outside );

          const int order = ulocal.order() + wlocal.order();
          const QuadratureRule &quadrature = QuadratureRules::rule( intersection.type(), order );
          const QuadratureIterator qend = quadrature.end();
          for( QuadratureIterator qit = quadrature.begin(); qit != qend; ++qit )
          {
            typename DomainLocalFunction::RangeType value;
            ulocal.evaluate( intersection.geometryInInside().global( qit->position() ), value );
            value *= qit->weight() * geometry.integrationElement( qit->position() );
            wlocal.axpy( intersection.geometryInOutside().global( qit->position() ), value );
          }
        }

        // apply inverse mass matrix to w (in place)

        LocalMassMatrix localMassMatrix( w.space(), 2*w.space().order() );

        const RangeIterator rend = w.space().end();
        for( RangeIterator rit = w.space().begin(); rit != rend; ++rit )
        {
          const typename RangeEntityPointer::Entity &entity = *rit;
          RangeLocalFunction wlocal = w.localFunction( entity );
          localMassMatrix.applyInverse( entity, wlocal );
        }
      }

    private:
      const GridPartGlueType &glue_;
    };

  } // namespace Ash

} // namespace Dune

#endif // #ifndef DUNE_ASH_GLUE_L2PROJECTION_HH
