/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

// vim: set expandtab ts=2 sw=2 sts=2:
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <SDL_image.h>

#include <dune/ash/image.hh>

namespace Dune
{

  namespace Ash
  {

    // SDLImage
    // --------

    struct SDLImage
    {
      SDLImage ();
      ~SDLImage () { IMG_Quit(); }

      SDL_Surface *load ( const std::string &file ) { return IMG_Load( file.c_str() ); }
    };



    // Single Instance of SDLImage
    // ---------------------------

    SDLImage sdlImage;



    // Implementation of SDLImage
    // --------------------------

    SDLImage::SDLImage ()
    {
      const int flags = IMG_INIT_JPG | IMG_INIT_PNG;
      if( IMG_Init( flags ) != flags  )
      {
        std::cerr << "SDL_image: " << IMG_GetError() << ".";
        abort();
      }

      SDL_version vCompiled;
      SDL_IMAGE_VERSION( &vCompiled );
      const SDL_version *vLinked = IMG_Linked_Version();
      if( vLinked->major != vCompiled.major )
        std::cerr << "Warning: A completely different version of SDL_image was used at compile time." << std::endl;
      if( vLinked->minor < vCompiled.minor )
        std::cerr << "Warning: An new version of SDL_image was used at compile time." << std::endl;
    }



    // loadImage
    // ---------

    array< GLsizei, 2 > loadImage ( const std::string &file, Texture &texture )
    {
      SDL_Surface *surface = sdlImage.load( file.c_str() );
      if( !surface )
        DUNE_THROW( Exception, "Unable to load image from '" << file << "'." );

      array< GLsizei, 2 > size;
      size[ 0 ] = surface->w;
      size[ 1 ] = surface->h;

      GLenum format = (surface->format->Amask ? GL_RGBA : GL_RGB);
      texture.load( GL_RGBA, size, 0, format, (const GLubyte *)surface->pixels );

      SDL_FreeSurface( surface );
      return size;
    }

  } // namespace Ash

} // namespace Dune
