/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

#ifndef IO_hh
#define IO_hh

#include <cstdlib>
#include <iostream>
#include <fstream>
#include <list>
#include <string>

#include <dune/fem/io/io.hh>

template< class MousePathType >
void writeField( const std::list< MousePathType > &data, const std::string &filename, bool forceWriting = false )
{
  typedef typename std::list< MousePathType >::const_iterator ListIteratorType;
  typedef typename MousePathType::const_iterator PathIteratorType;
  typedef typename MousePathType::value_type DataTuple;
  typedef typename DataTuple::first_type PositionType;
  typedef typename DataTuple::second_type DirectionType;

  if( !forceWriting && Dune::fileExists( filename ) )
  {
    std::cerr << "Error in writeField: '" << filename << "' already exists!\n";
    abort();
  }

  std::ofstream outstream( filename.c_str() );
  for( ListIteratorType it = data.begin(); it != data.end(); ++it )
  {
    for( PathIteratorType pit = it->begin(); pit != it->end(); ++pit )
    {
      const PositionType &position = pit->first;
      const PositionType &direction = pit->second;
      outstream << position[ 0 ] << " " << position[ 1 ] << "  ";
      outstream << direction[ 0 ] << " " << direction[ 1 ] << std::endl;
    }
    outstream << std::endl;
  }
  outstream.close();
}

template< class MousePathType >
void readField( std::list< MousePathType > &data, const std::string &filename, bool forceReading = false )
{
  typedef typename MousePathType :: iterator      PathIteratorType;
  typedef typename MousePathType :: value_type    DataTuple;
  typedef typename DataTuple :: first_type        PositionType;
  typedef typename DataTuple :: second_type       DirectionType;

  data.clear();

  std::ifstream instream( filename.c_str() );
  if( !instream )
  {
    std::cerr << "Error: Cannot open \"" << filename << "\"." << std::endl;
    if( !forceReading )
      abort();
    return;
  }

  std::string line;
  std::vector< double > row;
  MousePathType path;
  while( std::getline( instream, line ) )
  {
    row.clear();
    std::istringstream ss( line );

    if( line.empty() )
    {
      if( !path.empty() )
        data.push_back( path );
      path.clear();
    }
    else
    {
      double x;
      while( ss >> x )
        row.push_back( x );

      // assert that we read in 4 doubles
      if( row.size() != 4 )
      {
        std::cerr << "File \"" << filename << "\" is faulty." << std::endl;
        abort();
      }

      PositionType p;
      DirectionType d;
      p[0] = row[0];
      p[1] = row[1];
      d[0] = row[2];
      d[1] = row[3];

      // append data to MousePathType structure
      path.push_back( std::make_pair( p, d ) );
    }
  }
  if( !path.empty() )
    data.push_back( path );

  instream.close();
}

#endif // IO_hh

