/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

#ifndef DUNE_ASH_OPENGL_FUNCTOR_HH
#define DUNE_ASH_OPENGL_FUNCTOR_HH

#include <dune/ash/fvector.hh>
#include <dune/ash/opengl/opengl.hh>

namespace Dune
{

  namespace Ash
  {

    // Rectangle
    // ---------

    struct Rectangle
    {
      typedef Dune::FieldVector< GLfloat, 3 > Position;
      typedef Dune::FieldVector< GLfloat, 4 > Color;

      Rectangle ( const Position &upperLeft, const Position &lowerRight,
                  const Color &color = makeFieldVector( 1.0f, 1.0f, 1.0f, 1.0f ) )
      : upperLeft_( upperLeft ), lowerRight_( lowerRight ),
        color_( color )
      {}

      explicit Rectangle ( const Color &color = makeFieldVector( 1.0f, 1.0f, 1.0f, 1.0f ) )
      : upperLeft_( makeFieldVector( 0, 0, 0 ) ), lowerRight_( makeFieldVector( 1, 1, 0 ) ),
        color_( color )
      {}

      void operator() ()
      {
        glBegin( GL_QUADS );
        {
          glColor( color_ );
          glQuad( upperLeft_, lowerRight_ );
        }
        glEnd();
      }

    private:
      Position upperLeft_, lowerRight_;
      Color color_;
    };

 

    // RectangleOutline
    // ----------------

    struct RectangleOutline
    : public Rectangle
    {
      RectangleOutline ( const Position &upperLeft, const Position &lowerRight,
                         const Color &color = makeFieldVector( 1.0f, 1.0f, 1.0f, 1.0f ),
                         GLfloat lineWidth = 1.0f )
      : Rectangle( upperLeft, lowerRight, color ),
        lineWidth_( lineWidth )
      {}

      explicit RectangleOutline ( const Color &color, GLfloat lineWidth = 1.0f )
      : Rectangle( color ),
        lineWidth_( lineWidth )
      {}

      explicit RectangleOutline ( GLfloat lineWidth = 1.0f )
      : Rectangle(),
        lineWidth_( lineWidth )
      {}

      void operator() ()
      {
        glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
        glLineWidth( lineWidth_ );
        static_cast< Rectangle & >( *this )();
        glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
      }

    private:
      GLfloat lineWidth_;
    };



    // AbstractFunctor
    // ---------------

    struct AbstractFunctor
    {
      virtual ~AbstractFunctor () {}
      virtual void operator() () = 0;
    };



    // FunctorWrapper
    // --------------

    template< class Functor >
    struct FunctorWrapper
    : public AbstractFunctor
    {
      FunctorWrapper ( Functor functor )
      : functor_( functor )
      {}

      void operator() () { functor_(); }

    private:
      Functor functor_;
    };

  } // namespace Ash

} // namespace Dune

#endif // #ifndef DUNE_ASH_OPENGL_DISPLAYLIST_HH
