/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

#include <config.h>

#include <dune/ash/opengl/texture.hh>

namespace Dune
{

  namespace Ash
  {

    void Texture::load ( GLint internalFormat, const array< GLsizei, 2 > &size, GLint border, AbstractFunctor &functor )
    {
      // save view port and clear color
      FieldVector< GLfloat, 4 > saveClearColor;
      array< GLint, 2 > saveViewportPos;
      array< GLsizei, 2 > saveViewportSize;
      glGetClearColor( saveClearColor );
      glGetViewport( saveViewportPos, saveViewportSize );

      // save projection matrix
      glMatrixMode( GL_PROJECTION );
      glPushMatrix();

      setup( internalFormat, size, border );

      array< GLint, 2 > vpPos;
      array< GLsizei, 2 > vpSize;
      for( vpPos[ 1 ] = 0; vpPos[ 1 ] < size[ 1 ]; )
      {
        vpSize[ 1 ] = std::min( saveViewportSize[ 1 ], size[ 1 ] - vpPos[ 1 ] );

        const GLfloat top = GLfloat( vpPos[ 1 ] ) / GLfloat( size[ 1 ] );
        const GLfloat bottom = GLfloat( vpPos[ 1 ] + vpSize[ 1 ] ) / GLfloat( size[ 1 ] );

        for( vpPos[ 0 ] = 0; vpPos[ 0 ] < size[ 0 ] ; )
        {
          vpSize[ 0 ] = std::min( saveViewportSize[ 0 ], size[ 0 ] - vpPos[ 0 ] );

          const GLfloat left = GLfloat( vpPos[ 0 ] ) / GLfloat( size[ 0 ] );
          const GLfloat right = GLfloat( vpPos[ 0 ] + vpSize[ 0 ] ) / GLfloat( size[ 0 ] );

          // set view port and projection matrix
          glViewport( saveViewportPos, vpSize );
          glMatrixMode( GL_PROJECTION );
          glLoadIdentity();
          glOrtho( left, right, top, bottom, 0.0, 1.0 );


          // clear frame buffer and render
          glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );
          glClear( GL_COLOR_BUFFER_BIT );
          functor();

          // copy the frame buffer content into the texture
          select();
          glCopyTexSubImage2D( GL_TEXTURE_2D, 0, vpPos[ 0 ], vpPos[ 1 ], saveViewportPos[ 0 ], saveViewportPos[ 1 ], vpSize[ 0 ], vpSize[ 1 ] );
          DUNE_ASH_CHECK_GLERROR;

          vpPos[ 0 ] += vpSize[ 0 ];
        }
        vpPos[ 1 ] += vpSize[ 1 ];
      }

      // restore projection matrix
      glMatrixMode( GL_PROJECTION );
      glPopMatrix();

      // restore view port and clear color
      glViewport( saveViewportPos, saveViewportSize );
      glClearColor( saveClearColor );
      DUNE_ASH_CHECK_GLERROR;
    }

  } // namespace Ash

} // namespace Dune
