/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

#ifndef DUNE_ASH_OPENGL_TEXTURE_HH
#define DUNE_ASH_OPENGL_TEXTURE_HH

#include <GL/gl.h>

#include <dune/common/nullptr.hh>

#include <dune/ash/fvector.hh>
#include <dune/ash/opengl/error.hh>
#include <dune/ash/opengl/functor.hh>
#include <dune/ash/opengl/opengl.hh>

namespace Dune
{

  namespace Ash
  {

    // Texture
    // -------

    struct Texture
    {
      Texture () { glGenTextures( 1, &handle_ ); }
      ~Texture () { glDeleteTextures( 1, &handle_ ); }

      void select () const { glBindTexture( GL_TEXTURE_2D, handle_ ); }

      template< class T >
      void load ( GLint internalFormat, const array< GLsizei, 1 > &size, GLint border, GLenum format, const T *data )
      {
        glBindTexture( GL_TEXTURE_1D, handle_ );
        glTexParameteri( GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
        glTexParameteri( GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
        glTexImage1D( GL_TEXTURE_1D, 0, internalFormat, size[ 0 ], border, format, GLTypeIdentifier< T >::value, data );
        DUNE_ASH_CHECK_GLERROR;
      }

      template< class T >
      void load ( GLint internalFormat, const array< GLsizei, 2 > &size, GLint border, GLenum format, const T *data )
      {
        glBindTexture( GL_TEXTURE_2D, handle_ );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
        glTexImage2D( GL_TEXTURE_2D, 0, internalFormat, size[ 0 ], size[ 1 ], border, format, GLTypeIdentifier< T >::value, data );
        DUNE_ASH_CHECK_GLERROR;
      }

      template< class T >
      void load ( GLint internalFormat, const array< GLsizei, 3 > &size, GLint border, GLenum format, const T *data )
      {
        glBindTexture( GL_TEXTURE_3D, handle_ );
        glTexParameteri( GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
        glTexParameteri( GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
        glTexImage3D( GL_TEXTURE_3D, 0, internalFormat, size[ 0 ], size[ 1 ], size[ 2 ], border, format, GLTypeIdentifier< T >::value, data );
        DUNE_ASH_CHECK_GLERROR;
      }

      void load ( GLint internalFormat, const array< GLsizei, 2 > &size, GLint border, AbstractFunctor &functor );

      template< class Functor >
      void load ( GLint internalFormat, const array< GLsizei, 2 > &size, GLint border, Functor functor )
      {
        FunctorWrapper< Functor > functorWrapper( functor );
        load( internalFormat, size, border, static_cast< AbstractFunctor & >( functorWrapper ) );
      }

      void setup ( GLint internalFormat, const array< GLsizei, 1 > &size, GLint border )
      {
        glBindTexture( GL_TEXTURE_1D, handle_ );
        glTexParameteri( GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
        glTexParameteri( GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
        glTexImage1D( GL_TEXTURE_1D, 0, internalFormat, size[ 0 ], border, GL_ALPHA, GL_UNSIGNED_BYTE, nullptr );
        DUNE_ASH_CHECK_GLERROR;
      }

      void setup ( GLint internalFormat, const array< GLsizei, 2 > &size, GLint border )
      {
        glBindTexture( GL_TEXTURE_2D, handle_ );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
        glTexImage2D( GL_TEXTURE_2D, 0, internalFormat, size[ 0 ], size[ 1 ], border, GL_ALPHA, GL_UNSIGNED_BYTE, nullptr );
        DUNE_ASH_CHECK_GLERROR;
      }

      void setup ( GLint internalFormat, const array< GLsizei, 3 > &size, GLint border )
      {
        glBindTexture( GL_TEXTURE_3D, handle_ );
        glTexParameteri( GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
        glTexParameteri( GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
        glTexImage3D( GL_TEXTURE_3D, 0, internalFormat, size[ 0 ], size[ 1 ], size[ 2 ], border, GL_ALPHA, GL_UNSIGNED_BYTE, nullptr );
        DUNE_ASH_CHECK_GLERROR;
      }

    private:
      // forbid copying and assignment
      Texture ( const Texture & );
      const Texture &operator= ( const Texture & );

      GLuint handle_;
    };



    // BillBoard
    // ---------

    struct BillBoard
    {
      typedef Dune::FieldVector< GLfloat, 3 > Position;
      typedef Dune::FieldVector< GLfloat, 2 > TexPosition;
      typedef Dune::FieldVector< GLfloat, 4 > Color;

      BillBoard ( const Position &upperLeft, const Position &lowerRight,
                  const Texture &texture,
                  const TexPosition &texUpperLeft = makeFieldVector( 0, 0 ),
                  const TexPosition &texLowerRight = makeFieldVector( 1, 1 ),
                  const Color &color = makeFieldVector( 1, 1, 1, 1 ) )
      : upperLeft_( upperLeft ), lowerRight_( lowerRight ),
        texture_( texture ),
        texUpperLeft_( texUpperLeft ), texLowerRight_( texLowerRight ),
        color_( color )
      {}

      BillBoard ( const Position &upperLeft, const Position &lowerRight,
                  const Texture &texture, const Color &color )
      : upperLeft_( upperLeft ), lowerRight_( lowerRight ),
        texture_( texture ),
        texUpperLeft_( makeFieldVector( 0, 0 ) ), texLowerRight_( makeFieldVector( 1, 1 ) ),
        color_( color )
      {}

      BillBoard ( const Texture &texture,
                  const TexPosition &texUpperLeft = makeFieldVector( 0, 0 ),
                  const TexPosition &texLowerRight = makeFieldVector( 1, 1 ),
                  const Color &color = makeFieldVector( 1, 1, 1, 1 ) )
      : upperLeft_( makeFieldVector( 0, 0, 1 ) ), lowerRight_( makeFieldVector( 1, 1, 1 ) ),
        texture_( texture ),
        texUpperLeft_( texUpperLeft ), texLowerRight_( texLowerRight ),
        color_( color )
      {}

      BillBoard ( const Texture &texture, const Color &color )
      : upperLeft_( makeFieldVector( 0, 0, 1 ) ), lowerRight_( makeFieldVector( 1, 1, 1 ) ),
        texture_( texture ),
        texUpperLeft_( makeFieldVector( 0, 0 ) ), texLowerRight_( makeFieldVector( 1, 1 ) ),
        color_( color )
      {}

      void operator() ()
      {
        glEnable( GL_TEXTURE_2D );

        glColor( color_ );
        texture_.select();
        glBegin( GL_QUADS );
        {
          glTexCoord( texUpperLeft_[ 0 ], texUpperLeft_[ 1 ] );
          glVertex( upperLeft_[ 0 ], upperLeft_[ 1 ]);
          glTexCoord( texLowerRight_[ 0 ], texUpperLeft_[ 1 ] );
          glVertex( lowerRight_[ 0 ], upperLeft_[ 1 ] );
          glTexCoord( texLowerRight_[ 0 ], texLowerRight_[ 1 ] );
          glVertex( lowerRight_[ 0 ], lowerRight_[ 1 ] );
          glTexCoord( texUpperLeft_[ 0 ], texLowerRight_[ 1 ] );
          glVertex( upperLeft_[ 0 ], lowerRight_[ 1 ] );
        }
        glEnd();

        glDisable( GL_TEXTURE_2D );
      }

    private:
      Position upperLeft_, lowerRight_;
      const Texture &texture_;
      TexPosition texUpperLeft_, texLowerRight_;
      Color color_;
    };

  } // namespace Ash

} // namespace Dune

#endif // #ifndef DUNE_ASH_OPENGL_TEXTURE_HH
