/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

#include <config.h>

#include <dune/ash/text/fontmanager.hh>

namespace Dune
{

  namespace Ash
  {

    // Implementation of FontManager
    // -----------------------------

    void FontManager::setFont ( const FontId &fontId )
    {
      const FontMap::iterator pos = fontMap_.find( fontId );
      if( pos == fontMap_.end() )
      {
        stateStack_.top().font_ = loadFont( fontId, paths_ );
        if( !stateStack_.top().font_ )
        {
          if( !defaultFont_ )
            DUNE_THROW( InvalidStateException, "Unable to load default font '" << fontId.first << "'." );
          else
            std::cerr << "Font '" << fontId.first << "' not found, using default font instead." << std::endl;
          stateStack_.top().font_ = defaultFont_;
        }
        fontMap_.insert( std::make_pair( fontId, stateStack_.top().font_ ) );
      }
      else
        stateStack_.top().font_ = pos->second;
    }


    void FontManager::setup ()
    {
      while( !stateStack_.empty() )
        stateStack_.pop();

      stateStack_.push( FontState( nullptr, makeFieldVector( 0.0f, 0.0f, 0.0f, 1.0f ), 0 ) );

      defaultFont_ = nullptr;
      setFont( defaultFont() );
      defaultFont_ = stateStack_.top().font_;

      space_ = font().size( " " )[ 0 ];
      lineHeight_ = font().maxAscent() - font().maxDescent();

      stateStack_.top().style_ = Parameter::getValue< int >( "ash.fontmanager.defaultstyle", 0 );
      setLeading( 0.2f );
    }


    void FontManager::push () { stateStack_.push( stateStack_.top() ); }


    void FontManager::pop () 
    { 
      if ( !stateStack_.empty() )
        stateStack_.pop();
      if ( stateStack_.empty() )
        setup();  
    }

    void FontManager::reset () { setup(); }


    std::vector< std::string > FontManager::fontPaths ()
    {
      std::vector< std::string > fontPaths;
      std::string paths = Parameter::getValue< std::string >( "ash.fontpaths" );
      const std::size_t pathLen = paths.length();
      for( std::size_t begin = 0; begin < pathLen; )
      {
        std::size_t end = paths.find_first_of( ':', begin );
        if( end == paths.npos )
          end = pathLen;
        if( end - begin > 0 )
          fontPaths.push_back( paths.substr( begin, end - begin ) );
        begin = end+1;
      }
      return fontPaths;
    }


    TrueTypeFont *FontManager::loadFont ( const std::string &font, int ptSize,
                                          const std::vector< std::string > &paths )
    {
      for( std::vector< std::string >::const_iterator it = paths.begin(); it != paths.end(); ++it )
      {
        const std::string fileName = *it + "/" + font + ".ttf";
        if( Dune::fileExists( fileName ) )
          return new TrueTypeFont( fileName, ptSize );
      }
      return nullptr;
    }


    TrueTypeFont *FontManager::loadFont ( const FontId &fontId,
                                          const std::vector< std::string > &paths )
    {
      return loadFont( fontId.first, fontId.second, paths );
    }


    typename FontManager::FontId FontManager::defaultFont ()
    {
      const std::string font = Parameter::getValue< std::string >( "ash.fontmanager.defaultfont" );
      const int ptSize = Parameter::getValue< int >( "ash.fontmanager.defaultsize" );
      return FontId( font, ptSize );
    }

  } // namespace Ash

} // namespace Dune
