/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

#ifndef DUNE_ASH_TEXT_PARSER_HH
#define DUNE_ASH_TEXT_PARSER_HH

#include <iostream>
#include <list>
#include <map>
#include <vector>

#include <dune/common/exceptions.hh>
#include <dune/common/nullptr.hh>

#include <dune/fem/io/parameter.hh>

#include <dune/ash/text/box.hh>

namespace Dune
{

  namespace Ash
  {

    class TextParser
    {
      typedef TextParser This;

      struct CommandInterface;

      struct DefCommand;            // \def{macro}{substitution}
      struct UndefCommand;          // \undef{macro}
      struct IfDefCommand;          // \ifdef{macro}{then} or \ifdef{macro}{then}{else}
      struct IfEqCommand;           // \ifeq{macro}{macro}{then} or \ifeq{macro}{macro}{then}{else}
      struct AdvanceCommand;        // \advance{macro}
      struct InputCommand;          // \input{filename}

      struct NewLineCommand;        // \newline (or \\)
      struct HorizontalFillCommand; // \hfill
      struct ImageCommand;          // \image{width}{height}{filename}
      struct ButtonCommand;         // \button{eventId}{content}
      struct GlueCommand;           // \glue{content}
      struct MBoxCommand;           // \mbox{content}
      struct AbsoluteCommand;       // \absolute{x}{y}{content}
      struct ParBoxCommand;         // \parbox{width}{content}
      struct ColorCommand;          // \color{literal} or \color{r,g,b}
      struct FontCommand;           // \font{filename}{size}
      struct BoldCommand;           // \bf
      struct ItalicCommand;         // \it
      struct LeadingCommand;        // \leading{factor}

      typedef std::map< std::string, CommandInterface * > CommandMap;

    public:
      typedef std::list< BoxInterface * > BoxList;
      typedef std::map< std::string, std::string > MacroMap;

      TextParser ();
      explicit TextParser ( const std::string &textPath );

    private:
      // prohibit copying and assignment
      TextParser ( const This & );
      const This &operator= ( const This & );

    public:
      ~TextParser ();

      BoxList operator() ( std::istream &input ) const;
      BoxList operator() ( std::istream &input, MacroMap &macros ) const;
      void operator() ( std::istream &input, MacroMap &macros, BoxList &boxes ) const;

    private:
      template< class T >
      static T parseValue ( const std::string &s )
      {
        T value;
        if( !Dune::ParameterParser< T >::parse( s, value ) )
          DUNE_THROW( InvalidStateException, "Invalid Parameter (" << s << ")." );
        return value;
      }

      void setupCommands ();

      static int parseCommand ( std::istream &input, std::string &command, std::vector< std::string > &args );

      std::string textPath ( const std::string &fileName ) const
      {
        return (textPath_.empty() ? fileName : textPath_ + "/" + fileName);
      }

      std::string textPath_;
      CommandMap commands_;
    };

  } // namespace Ash

} // namespace Dune

#endif // #ifndef DUNE_ASH_TEXT_PARSER_HH
