/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

#ifndef DUNE_ASH_TEXTURES_GRIDFUNCTOR_HH
#define DUNE_ASH_TEXTURES_GRIDFUNCTOR_HH


#include <dune/ash/fvector.hh>
#include <dune/ash/opengl/functor.hh>
#include <dune/ash/opengl/opengl.hh>


namespace Dune
{

  namespace Ash
  {

    // GridFunctor
    // -----

    template< class Functor >
    struct GridFunctor
    {

      typedef FieldVector< GLfloat, 3 > Position;
      typedef FieldVector< GLfloat, 4 > Color;
      
      explicit GridFunctor( Functor functor, size_t level = 1, const Color &color = Color( 1.0f ), const Position &offset = Position( 0 ) )
      : functor_( functor ),
        border_( RectangleOutline( color ) ),
        offset_( offset ),
        color_( color ),
        level_( level )
      {}

      void operator() ()
      {
        glColor( color_ );

        for( std::size_t i = 0; i < level_*level_; ++i )
        {
          const GLfloat height = 1.0 / level_;
          const GLfloat width = 1.0 / level_;
          const Position position = makeFieldVector( ( width * ( i % level_ ) ), height * ( i / level_ ), 0.0f  );

          glMatrixMode( GL_MODELVIEW );
          glPushMatrix();
          glTranslate( position );
          glScale( width, height, 1.0f );
          glTranslate( offset_ );
          functor_();
          glPopMatrix();
        }

        border_();
      }

    private:
      Functor functor_;
      RectangleOutline border_;
      const Position offset_;
      const Color color_;
      const size_t level_;
    };

    template< class Functor >
    struct DeformFunctor
    {
      typedef FieldVector< GLfloat, 3 > Position;

      explicit DeformFunctor( Functor functor, const Position &topLeft = Position( 0.0f), const Position &bottomRight = makeFieldVector( 1.0f, 1.0f, 1.0f ) )
      : functor_( functor ),
        topLeft_( topLeft ),
        bottomRight_( bottomRight )
      {}

      void operator() ()
      {
        glMatrixMode( GL_MODELVIEW );
        glPushMatrix();
        glTranslate( topLeft_ );
        glScale( bottomRight_[ 0 ] - topLeft_[ 0 ], bottomRight_[ 1 ] - topLeft_[ 1 ], 1.0f );
        functor_();
        glPopMatrix();
      }

    private:
      Functor functor_;
      const Position topLeft_;
      const Position bottomRight_;
    };
    

  } // namespace Ash

} // namespace Dune


#endif //#ifndef DUNE_ASH_TEXTURES_GRIDFUNCTOR_HH


