/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

// vim: set expandtab ts=2 sw=2 sts=2: 
#ifndef DISCRETEFUNCTIONVISUALIZER_HH
#define DISCRETEFUNCTIONVISUALIZER_HH

#include <cassert>

#include <dune/grid/spgrid.hh>

#include <dune/fem/function/adaptivefunction.hh>
#include <dune/fem/gridpart/levelgridpart.hh>
#include <dune/fem/quadrature/cachingquadrature.hh>
#include <dune/fem/space/fvspace.hh>

#include <dune/ash/glue/gridpart.hh>
#include <dune/ash/glue/l2projection.hh>
#include <dune/ash/opengl/texture.hh>

namespace Dune
{

  namespace Ash
  {

    // DiscreteFunctionVisualizer
    // --------------------------

    template< class DiscreteFunction >
    class DiscreteFunctionVisualizer
    {
      typedef DiscreteFunctionVisualizer< DiscreteFunction > ThisType;

    public:
      typedef DiscreteFunction DiscreteFunctionType;

      typedef typename DiscreteFunctionType::GridPartType GridPartType;
      typedef typename DiscreteFunctionType::FunctionSpaceType FunctionSpaceType;

      typedef typename GridPartType::ctype ctype;

      static const int dimension = GridPartType::dimension;

    private:
      typedef SPGrid< ctype, dimension > RasterGridType;
      typedef Fem::LevelGridPart< RasterGridType > RasterGridPartType;
      typedef Fem::FiniteVolumeSpace< FunctionSpaceType, RasterGridPartType, 0 > RasterSpaceType;
      typedef Fem::AdaptiveDiscreteFunction< RasterSpaceType > RasterFunctionType;

      typedef typename RasterGridPartType::template Codim< 0 >::IteratorType RasterIteratorType;
      typedef typename RasterGridPartType::template Codim< 0 >::EntityType RasterEntityType;
      typedef typename RasterGridPartType::IndexSetType RasterIndexSetType;

      typedef typename RasterFunctionType::LocalFunctionType LocalRasterFunctionType;

    public:
      DiscreteFunctionVisualizer ( const DiscreteFunctionType &discreteFunction,
                                   const array< int, dimension > &rasterCells )
      : rasterCells_( rasterCells ),
        rasterGrid_( discreteFunction.gridPart().grid().domain(), rasterCells_ ),
        rasterGridPart_( rasterGrid_, 0 ),
        rasterSpace_( rasterGridPart_ ),
        rasterFunction_( "raster function", rasterSpace_ )
      {
        typedef GridPartGlue< GridPartType, RasterGridPartType > Glue;
        typedef GluedL2Projection< Glue > Projection;

        Glue glue( discreteFunction.gridPart(), rasterGridPart_ );
        Projection projection( glue );
        projection( discreteFunction, rasterFunction_ );
      }

      template< class ColorMapper >
      void loadTexture ( const ColorMapper &colorMapper, Texture &texture ) const
      {
        const RasterIndexSetType &rasterIndexSet = rasterGridPart_.indexSet();
        typename ColorMapper::Color *texels = new typename ColorMapper::Color[ rasterIndexSet.size( 0 ) ];

        const RasterIteratorType rend = rasterSpace_.end();
        for( RasterIteratorType rit = rasterSpace_.begin(); rit != rend; ++rit )
        {
          const RasterEntityType &entity = *rit;
          LocalRasterFunctionType localFunction = rasterFunction_.localFunction( entity );
          typename FunctionSpaceType::RangeType value;
          Fem::CachingQuadrature< RasterGridPartType, 0 > quadrature( entity, 0 );
          assert( quadrature.nop() == 1 );
          localFunction.evaluate( quadrature.point( 0 ), value );
          texels[ rasterIndexSet.index( entity ) ] = colorMapper.map( value );
        }

        texture.load( colorMapper.format(), rasterCells_, 0, colorMapper.format(), texels );
        delete texels;
      }

    private:
      // prohibit copying and assignment
      DiscreteFunctionVisualizer ( const ThisType & );
      const ThisType &operator= ( const ThisType & );

      array< int, 2 > rasterCells_;
      RasterGridType rasterGrid_;
      RasterGridPartType rasterGridPart_;
      RasterSpaceType rasterSpace_;
      RasterFunctionType rasterFunction_;
    };

  } // namespace Ash

} // namespace Dune

#endif // #ifndef DISCRETEFUNCTIONVISUALIZER_HH
