/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

// vim: set expandtab ts=2 sw=2 sts=2:
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <dune/common/nullptr.hh>

#include <dune/ash/opengl/opengl.hh>
#include <dune/ash/widget/collection.hh>

namespace Dune
{

  namespace Ash
  {

    // Implementation of CollectionWidget
    // ----------------------------------

    CollectionWidget::CollectionWidget ()
    : childList_(),
      active_( childList_.end() )
    {}


    void CollectionWidget::draw ()
    {
      for( ChildList::const_reverse_iterator it = childList_.rbegin(); it != childList_.rend(); ++it )
      {
        glMatrixMode( GL_MODELVIEW );
        glPushMatrix();

        glTranslate( it->origin );
        glScale( it->size );
        it->widget->draw();

        glPopMatrix();
      }
    }


    void CollectionWidget
      ::mouseMove ( const Position &x, const Position &dx, Ticks ticks )
    {
      if( active_ == childList_.end() )
      {
        ChildList::const_iterator pos = findChild( x );
        if( pos != childList_.end() )
          pos->widget->mouseMove( pos->position( x ), dx, ticks );
      }
      else
        active_->widget->mouseMove( active_->position( x ), dx, ticks );
    }


    void CollectionWidget
      ::mouseButtonDown ( const Position &x, Ticks ticks )
    {
      if( active_ == childList_.end() )
        active_ = findChild( x );
      if( active_ != childList_.end() )
        active_->widget->mouseButtonDown( active_->position( x ), ticks );
    }


    void CollectionWidget
      ::mouseButtonUp ( const Position &x, Ticks ticks )
    {
      if( active_ == childList_.end() )
      {
        ChildList::const_iterator pos = findChild( x );
        if( pos != childList_.end() )
          pos->widget->mouseButtonUp( pos->position( x ), ticks );
      }
      else
        active_->widget->mouseButtonUp( active_->position( x ), ticks );
      active_ = childList_.end();
    }


    void CollectionWidget::tick ( Ticks ticks )
    {
      for( ChildList::const_iterator it = childList_.begin(); it != childList_.end(); ++it )
        it->widget->tick( ticks );
    }


    void CollectionWidget::addChild ( const Position &origin, const Position &size, Widget *widget )
    {
      childList_.push_front( Child( origin, size, widget ) );
    }


    void CollectionWidget::removeChild ( Widget *widget )
    {
      for( ChildList::iterator it = childList_.begin(); it != childList_.end(); ++it )
      {
        if( it->widget != widget )
          continue;

        childList_.erase( it );
        return;
      }
    }


    CollectionWidget::ChildList::const_iterator
    CollectionWidget::findChild ( const Position &x ) const
    {
      for( ChildList::const_iterator it = childList_.begin(); it != childList_.end(); ++it )
      {
        if( it->contains( x ) )
          return it;
      }
      return childList_.end();
    }

  } // namespace Ash

} // namespace Dune
