/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

// vim: set expandtab ts=2 sw=2 sts=2:
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <fstream>

#include <dune/fem/io/parameter.hh>

#include <dune/ash/opengl/displaylist.hh>
#include <dune/ash/text/parser.hh>
#include <dune/ash/widget/floatingtext.hh>

namespace Dune
{

  namespace Ash
  {

    FloatingTextWidget::FloatingTextWidget ( GLfloat dpu, const FieldVector< GLfloat, 2 > &size, const std::string &textPath, std::istream &input )
    : fontManager_( dpu ),
      size_( size ),
      background_( Parameter::getValue< Color >( "ash.text.background", makeFieldVector( 1, 1, 1 ) ) ),
      parBox_( nullptr ),
      activeBox_( activeBoxes_.end() )
    {
      setText( textPath, input );
    }

    FloatingTextWidget::FloatingTextWidget ( GLfloat dpu, const FieldVector< GLfloat, 2 > &size, std::istream &input )
    : fontManager_( dpu ),
      size_( size ),
      background_( Parameter::getValue< Color >( "ash.text.background", makeFieldVector( 1, 1, 1 ) ) ),
      parBox_( nullptr ),
      activeBox_( activeBoxes_.end() )
    {
      setText( input );
    }

    FloatingTextWidget::FloatingTextWidget ( GLfloat dpu, const FieldVector< GLfloat, 2 > &size, const std::string &textPath, const std::string &fileName )
    : fontManager_( dpu ),
      size_( size ),
      background_( Parameter::getValue< Color >( "ash.text.background", makeFieldVector( 1, 1, 1 ) ) ),
      parBox_( nullptr ),
      activeBox_( activeBoxes_.end() )
    {
      setText( textPath, fileName );
    }

    FloatingTextWidget::FloatingTextWidget ( GLfloat dpu, const FieldVector< GLfloat, 2 > &size, const std::string &fileName )
    : fontManager_( dpu ),
      size_( size ),
      background_( Parameter::getValue< Color >( "ash.text.background", makeFieldVector( 1, 1, 1 ) ) ),
      parBox_( nullptr ),
      activeBox_( activeBoxes_.end() )
    {
      setText( fileName );
    }

    FloatingTextWidget::FloatingTextWidget ( GLfloat dpu, const FieldVector< GLfloat, 2 > &size )
    : fontManager_( dpu ),
      size_( size ),
      background_( Parameter::getValue< Color >( "ash.text.background", makeFieldVector( 1, 1, 1 ) ) ),
      parBox_( nullptr ),
      activeBox_( activeBoxes_.end() )
    {}

    FloatingTextWidget::FloatingTextWidget ( const FieldVector< GLfloat, 2 > &dpu, const FieldVector< GLfloat, 2 > &size, const std::string &textPath, std::istream &input )
    : fontManager_( dpu ),
      size_( size ),
      background_( Parameter::getValue< Color >( "ash.text.background", makeFieldVector( 1, 1, 1 ) ) ),
      parBox_( nullptr ),
      activeBox_( activeBoxes_.end() )
    {
      setText( textPath, input );
    }

    FloatingTextWidget::FloatingTextWidget ( const FieldVector< GLfloat, 2 > &dpu, const FieldVector< GLfloat, 2 > &size, std::istream &input )
    : fontManager_( dpu ),
      size_( size ),
      background_( Parameter::getValue< Color >( "ash.text.background", makeFieldVector( 1, 1, 1 ) ) ),
      parBox_( nullptr ),
      activeBox_( activeBoxes_.end() )
    {
      setText( input );
    }

    FloatingTextWidget::FloatingTextWidget ( const FieldVector< GLfloat, 2 > &dpu, const FieldVector< GLfloat, 2 > &size, const std::string &textPath, const std::string &fileName )
    : fontManager_( dpu ),
      size_( size ),
      background_( Parameter::getValue< Color >( "ash.text.background", makeFieldVector( 1, 1, 1 ) ) ),
      parBox_( nullptr ),
      activeBox_( activeBoxes_.end() )
    {
      setText( textPath, fileName );
    }

    FloatingTextWidget::FloatingTextWidget ( const FieldVector< GLfloat, 2 > &dpu, const FieldVector< GLfloat, 2 > &size, const std::string &fileName )
    : fontManager_( dpu ),
      size_( size ),
      background_( Parameter::getValue< Color >( "ash.text.background", makeFieldVector( 1, 1, 1 ) ) ),
      parBox_( nullptr ),
      activeBox_( activeBoxes_.end() )
    {
      setText( fileName );
    }

    FloatingTextWidget::FloatingTextWidget ( const FieldVector< GLfloat, 2 > &dpu, const FieldVector< GLfloat, 2 > &size )
    : fontManager_( dpu ),
      size_( size ),
      background_( Parameter::getValue< Color >( "ash.text.background", makeFieldVector( 1, 1, 1 ) ) ),
      parBox_( nullptr ),
      activeBox_( activeBoxes_.end() )
    {}


    FloatingTextWidget::~FloatingTextWidget ()
    {
      if( parBox_ )
       delete parBox_; 
    }


    void FloatingTextWidget::setText ( const std::string &textPath, std::istream &input )
    {
      if( parBox_ )
        delete parBox_;
      parBox_ = new ParBox( size_[ 0 ], TextParser( textPath )( input ) );
      setupText();
    }

    void FloatingTextWidget::setText ( std::istream &input )
    {
      if( parBox_ )
        delete parBox_;
      parBox_ = new ParBox( size_[ 0 ], TextParser()( input ) );
      setupText();
    }

    void FloatingTextWidget::setText ( const std::string &textPath, const std::string &fileName )
    {
      std::ifstream input( (textPath + "/" + fileName).c_str() );
      setText( textPath, input );
    }

    void FloatingTextWidget::setText ( const std::string &fileName )
    {
      std::ifstream input( fileName.c_str() );
      setText( input );
    }


    void FloatingTextWidget::setupText ()
    {
      fontManager_.reset();
      parBox_->align( fontManager_ );

      array< GLsizei, 2 > rasterSize;
      rasterSize[ 0 ] = GLsizei( std::round( fontManager_.dotsPerUnit()[ 0 ] * size_[ 0 ] ) );
      rasterSize[ 1 ] = GLsizei( std::round( fontManager_.dotsPerUnit()[ 1 ] * size_[ 1 ] ) );

      glMatrixMode( GL_MODELVIEW );
      glPushMatrix();
      glLoadIdentity();
      glScale(  1.0f / size_[ 0 ], 1.0f / size_[ 1 ], 1.0f );

      texture_.load( GL_RGBA, rasterSize, 0, RenderParBox( *parBox_ ) );
      
      glMatrixMode( GL_MODELVIEW );
      glPopMatrix();

      activeBoxes_.clear();
      parBox_->gatherActive( makeFieldVector( 0.0f, parBox_->ascent() ), activeBoxes_ );
      activeBox_ = activeBoxes_.end();

      displayList_.compile( BillBoard( texture_ ) );
    }


    Event *FloatingTextWidget::setEvent ( const std::string &id, Event *event )
    {
      const EventMap::iterator pos = events_.find( id );
      if( pos != events_.end() )
      {
        std::swap( event, pos->second );
        if( !pos->second )
          events_.erase( pos );
        return event;
      }
      else
      {
        if( event )
          events_.insert( std::make_pair( id, event ) );
        return nullptr;
      }
    }


    void FloatingTextWidget::deleteEvents ()
    {
      for( EventMap::iterator it = events_.begin(); it != events_.end(); ++it )
        delete it->second;
      events_.clear();
    }


    void FloatingTextWidget::draw ()
    {
      Rectangle( makeFieldVector( background_[ 0 ], background_[ 1 ], background_[ 2 ], 1.0f ) )();
      if( parBox_ )
        displayList_();
    }


    void FloatingTextWidget::mouseMove ( const Position &x, const Position &dx, Ticks ticks )
    {}


    void FloatingTextWidget::mouseButtonDown ( const Position &x, Ticks ticks )
    {
      activeBox_ = activeBoxes_.end();
      const Position position = makeFieldVector( x[ 0 ] * size_[ 0 ], x[ 1 ] * size_[ 1 ] );
      for( ActiveBoxList::const_iterator it = activeBoxes_.begin(); it != activeBoxes_.end(); ++it )
      {
        if( it->contains( position ) )
          activeBox_ = it;
      }
    }


    void FloatingTextWidget::mouseButtonUp ( const Position &x, Ticks ticks )
    {
      const Position position = makeFieldVector( x[ 0 ] * size_[ 0 ], x[ 1 ] * size_[ 1 ] );
      if( (activeBox_ != activeBoxes_.end()) && activeBox_->contains( position ) )
      {
        const EventMap::const_iterator pos = events_.find( activeBox_->eventId() );
        if( pos != events_.end() )
          pos->second->executeEvent();
        else
          std::cerr << "Error: Event '" << activeBox_->eventId() << "' undefined." << std::endl;
      }
      activeBox_ = activeBoxes_.end();
    }

  } // namespace Ash

} // namespace Dune
