/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

// vim: set expandtab ts=2 sw=2 sts=2:
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <dune/common/nullptr.hh>

#include <dune/ash/opengl/opengl.hh>
#include <dune/ash/widget/grid.hh>

namespace Dune
{

  namespace Ash
  {

    // Implementation of GridWidget
    // ----------------------------

    GridWidget::GridWidget ( std::size_t horizontal, std::size_t vertical,
                             const Color &background, const Position &border )
    : border_( border ),
      child_( horizontal * vertical, nullptr ),
      active_( nullptr ),
      rect_( makeFieldVector( background[ 0 ], background[ 1 ], background[ 2 ], 1.0f ) )
    {
      size_[ 0 ] = horizontal;
      size_[ 1 ] = vertical;
      for( int i = 0; i < 2; ++i )
        diff_[ i ] =  (1.0 - border_[ i ]) / size_[ i ];
    }


    void GridWidget::draw ()
    {
      rect_();

      Index index;
      for( index[ 0 ] = 0; index[ 0 ] < size_[ 0 ] ;++index[ 0 ] ) 
      {
        for( index[ 1 ] = 0; index[ 1 ] < size_[ 1 ]; ++index[ 1 ] )
          draw( getChild( index ), origin( index ), size( index ) );
      }
    }


    void GridWidget::mouseMove ( const Position &x, const Position &dx, Ticks ticks )
    {
      if( active_ )
        active_->mouseMove( childPosition( activeIndex_, x ), dx, ticks );
      else  
      {
        Index index;
        Widget *child = getChild( x, index );
        if( child )
          child->mouseMove( childPosition( index, x ), dx, ticks );
      }
    }


    void GridWidget::mouseButtonDown ( const Position &x, Ticks ticks )
    {
      if( !active_ )
        active_ = getChild( x, activeIndex_ );
      if( active_ )
        active_->mouseButtonDown( childPosition( activeIndex_, x ), ticks );
    }


    void GridWidget::mouseButtonUp ( const Position &x, Ticks ticks )
    {
      if( active_ )
      {
        active_->mouseButtonUp( childPosition( activeIndex_, x ), ticks );
        active_ = nullptr;
      }
      else
      {
        Index index;
        Widget *child = getChild( x, index );
        if( child )
          child->mouseButtonUp( childPosition( index, x ), ticks );
      }
    }


    void GridWidget::draw ( Widget *child, const Position &origin, const Position &size )
    {
      if( !child )
        return;

      glMatrixMode( GL_MODELVIEW );
      glPushMatrix();

      FieldVector< GLdouble, 3 > t( 0.0 );
      t[ 0 ] = origin[ 0 ];
      t[ 1 ] = origin[ 1 ];
      glTranslate( t );

      FieldVector< GLdouble, 3 > s( 1.0 );
      s[ 0 ] = size[ 0 ];
      s[ 1 ] = size[ 1 ];
      glScale( s );

      child->draw();

      glPopMatrix();
    }


    void GridWidget::tick ( Ticks ticks )
    {
      Index index;
      for( index[ 0 ] = 0; index[ 0 ] < size_[ 0 ] ;++index[ 0 ] ) 
      {
        for( index[ 1 ] = 0; index[ 1 ] < size_[ 1 ]; ++index[ 1 ] )
        {
          Widget *child = getChild( index );
          if ( child )
            child->tick( ticks );
        }
      }
    }


    Widget *GridWidget::getChild ( const Position &x, Index &index )
    {
      for( int i = 0; i < 2; ++i )
      {
        index[ i ] = std::floor( x[ i ] / diff_[ i ] );
        if( (index[ i ] >= size_[ i ]) || (x[ i ] - index[ i ]*diff_[ i ] < border_[ i ]) )
          return nullptr;
      }
      return getChild( index );
    }


    typename GridWidget::Position
    GridWidget::childPosition ( const Index &index, const Position &x ) const
    {
      Position xChild( x );
      const Position o = origin( index );
      const Position s = size( index );
      for( int i = 0; i < 2; ++i )
        xChild[ i ] = (x[ i ] - o[ i ]) / s[ i ];
      return xChild;
    }


    typename GridWidget::Position
    GridWidget::origin ( const Index &index ) const
    {
      return makeFieldVector( index[ 0 ] * diff_[ 0 ], index[ 1 ] * diff_[ 1 ] ) + border_;
    }


    typename GridWidget::Position
    GridWidget::size ( const Index &index ) const
    {
      return diff_ - border_;
    }

  } // namespace Ash

} // namespace Dune
