/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

#ifndef DUNE_ASH_WIDGET_GRID_HH
#define DUNE_ASH_WIDGET_GRID_HH

#include <vector>

#include <dune/common/array.hh>

#include <dune/ash/fvector.hh>
#include <dune/ash/opengl/functor.hh>
#include <dune/ash/opengl/opengl.hh>
#include <dune/ash/widget/widget.hh>

namespace Dune
{

  namespace Ash
  {

    // GridWidget
    // ----------

    class GridWidget
    : public Widget
    {
      typedef GridWidget This;
      typedef Widget Base;

    public:
      typedef FieldVector< GLfloat, 3 > Color;
      typedef array< std::size_t, 2 > Index;

      GridWidget ( std::size_t horizontal, std::size_t vertical,
                   const Color &background = Color( 1 ), const Position &border = Position( 0 ) );

      void draw ();

      void mouseMove ( const Position &x, const Position &dx, Ticks ticks );

      void mouseButtonDown ( const Position &x, Ticks ticks );
      void mouseButtonUp ( const Position &x, Ticks ticks );

      Widget *getChild ( const Index &index ) const { return getChild( index[ 0 ], index[ 1 ] ); }
      Widget *getChild ( std::size_t h, std::size_t v ) const;
      Widget *setChild ( const Index &index, Widget *child ) { return setChild( index[ 0 ], index[ 1 ], child ); }
      Widget *setChild ( std::size_t h, std::size_t v, Widget *child );
     
      Index size () const { return size_; }

      void tick ( Ticks ticks );

    private:
      void draw ( Widget *child, const Position &origin, const Position &size );

      Widget *getChild ( const Position &x, Index &index );

      Position childPosition ( const Index &index, const Position &x ) const;
      Position origin ( const Index &index ) const;
      Position size ( const Index &index ) const;

      Index size_;
      Position border_, diff_;
      std::vector< Widget * >  child_;
      Widget *active_;
      Index activeIndex_;
      Rectangle rect_;
    };



    // Implementation of GridWidget
    // ----------------------------

    inline Widget *
    GridWidget::getChild ( std::size_t h, std::size_t v ) const
    {
      assert( h < size_[ 0 ] );
      assert( v < size_[ 1 ] );
      return child_[ h * size_[ 1 ] + v ];
    }


    inline Widget *
    GridWidget::setChild ( std::size_t h, std::size_t v, Widget *child )
    {
      assert( h < size_[ 0 ] );
      assert( v < size_[ 1 ] );
      std::swap( child_[ h * size_[ 1 ] + v ], child );
      return child;
    }

  } // namespace Ash

} // namespace Dune

#endif // #ifndef DUNE_ASH_WIDGET_GRID_HH
