/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

// vim: set expandtab ts=2 sw=2 sts=2:
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <dune/fem/io/parameter.hh>

#include <dune/ash/widget/slider.hh>

namespace Dune
{

  namespace Ash
  {

    // Implementation of SliderWidget
    // ------------------------------

    template< Orientation orientation >
    SliderWidget< orientation >::SliderWidget ( const Range &range, const Position &border, GLfloat barWidth )
    : range_( range ),
      active_( range ),
      position_( range.first ),
      grabbed_( false ),
      border_( border ),
      barWidth_( barWidth ),
      colorBackground_( Parameter::getValue< Color >( "ash.slider.background", makeFieldVector( 1, 1, 1 ) ) ),
      colorForeground_( Parameter::getValue< Color >( "ash.slider.foreground", makeFieldVector( 0, 0, 0 ) ) ),
      colorInactive_( Parameter::getValue< Color >( "ash.slider.inactive", makeFieldVector( 0.8f, 0.8f, 0.8f ) ) ),
      colorActive_( Parameter::getValue< Color >( "ash.slider.active", makeFieldVector( 0.7f, 0.7f, 1.0f ) ) )
    {}


    template< Orientation orientation >
    void SliderWidget< orientation >::draw ()
    {
      typedef FieldVector< GLfloat, 2 > Coordinate;

      const Coordinate width = makeFieldVector( 1.0f - 2.0f*border_[ 0 ], 1.0f - 2.0f*border_[ 1 ] );
      const GLfloat tic = width[ direction ] / GLfloat( range_.second - range_.first );

      const GLfloat areathickness = 0.15f;
      const GLfloat arrowdist = 1.5f*barWidth_*tic;
      const GLfloat arrowwidth = 2.0f*barWidth_*tic;
      const GLfloat arrowheight = 0.125f;

      glBegin( GL_QUADS );
      {
        glColor( colorBackground_ );
        glQuad( makeFieldVector( 0.0f, 0.0f ), makeFieldVector( 1.0f, 1.0f ) );

        Coordinate left = makeFieldVector( 0.5f - areathickness*width[ 0 ], 0.5f - areathickness*width[ 1 ] );
        Coordinate right = makeFieldVector( 0.5f + areathickness*width[ 0 ], 0.5f + areathickness*width[ 1 ] );

        left[ direction ] = border_[ direction ];
        right[ direction ] = 1.0f - border_[ direction ];

        glColor( colorInactive_ );
        glQuad( left, right );

        left[ direction ] += (active_.first - range_.first)*tic;
        right[ direction ] -= (range_.second - active_.second)*tic;

        glColor( colorActive_ );
        glQuad( left, right );

        left = makeFieldVector( border_[ 0 ], border_[ 1 ] );
        left[ direction ] = border_[ direction ] + (position_ - barWidth_)*tic;

        right = makeFieldVector( 1.0f - border_[ 0 ], 1.0f - border_[ 1 ] );
        right[ direction ] = border_[ direction ] + (position_ + barWidth_)*tic;

        glColor( colorForeground_ );
        glQuad( left, right );
      }
      glEnd();

      glBegin( GL_TRIANGLES );
      {
        glColor( colorForeground_ );

        Coordinate left = makeFieldVector( 0.5f - arrowheight*width[ 0 ], 0.5f - arrowheight*width[ 1 ] );
        left[ direction ] = border_[ direction ] + position_*tic - arrowdist;

        Coordinate right = makeFieldVector( 0.5f + arrowheight*width[ 0 ], 0.5f + arrowheight*width[ 1 ] );
        right[ direction ] = border_[ direction ] + position_*tic + arrowdist;

        if( position_ > active_.first )
        {
          Coordinate x;

          glVertex( left );

          x = makeFieldVector( 0.5f, 0.5f );
          x[ direction ] = left[ direction ] - arrowwidth;
          glVertex( x );

          x = right;
          x[ direction ] = left[ direction ];
          glVertex( x );
        }

        if( position_ < active_.second )
        {
          Coordinate x;

          x = left;
          x[ direction ] = right[ direction ];
          glVertex( x );

          x = makeFieldVector( 0.5f, 0.5f );
          x[ direction ] = right[ direction ] + arrowwidth;
          glVertex( x );

          glVertex( right );
        }
      }
      glEnd();
    }


    template< Orientation orientation >
    void SliderWidget< orientation >::mouseMove ( const Position &x, const Position &dx, Ticks ticks )
    {
      if( grabbed_ )
        followMouse( x );
    }


    template< Orientation orientation >
    void SliderWidget< orientation >::mouseButtonDown ( const Position &x, Ticks ticks )
    {
      grabbed_ = true;
      followMouse( x );
    }


    template< Orientation orientation >
    void SliderWidget< orientation >::mouseButtonUp ( const Position &x, Ticks ticks )
    {
      if( grabbed_ )
        followMouse( x );
      grabbed_ = false;
    }


    template< Orientation orientation >
    void SliderWidget< orientation >::followMouse ( const Position &x )
    {
      const GLfloat width = 1.0f - 2.0f*border_[ direction ];
      const GLfloat tic = width / GLfloat( range_.second - range_.first );
      if( x[ direction ] <= border_[ direction ] )
        position_ = active_.first;
      else if( x[ direction ] >= 1.0f - border_[ direction ] )
        position_ = active_.second;
      else
        position_ = clamp( std::size_t( (x[ direction ] + 0.5f*tic - border_[ direction ]) / tic ) );
    }


    template< Orientation orientation >
    std::size_t SliderWidget< orientation >::clamp ( std::size_t position ) const
    {
      return std::max( active_.first, std::min( active_.second, position ) );
    }


    template< Orientation orientation >
    void SliderWidget< orientation >::setActiveRange ( const Range &range )
    {
      active_.first = std::max( range_.first, range.first );
      active_.second = std::min( range_.second, range.second );
      if( active_.first > active_.second )
        std::swap( active_.first, active_.second );
      position_ = clamp( position_ );
    }


    // Template Instantiation
    // ----------------------

    template class SliderWidget< Horizontal >;
    template class SliderWidget< Vertical >;

  } // namespace Ash

} // namespace Dune
