/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

#include <config.h>

#include <iostream>
#include <string>

#include <dune/fem/misc/l2error.hh>
#include <dune/fem/operator/projection/l2projection.hh>
#include <dune/fem/gridpart/common/gridpart.hh>
#include <dune/fem/solver/odesolver.hh>

#include "../common/discretizationtraits.hh"
#include "../common/filemanager.hh"
#include "stepper.hh"


// main 
// ----

int main ( int argc, char **argv )
try
{
  // Initialize MPI (always do this even if you are not using MPI)
  Dune::MPIManager::initialize( argc, argv );

  // Initialize parameter data base
  Dune::Parameter::append( argc, argv );
  for( int i = 1; i < argc; ++i )
    Dune::Parameter::append( argv[ i ] );

  // load the parameter files required by ash
  Ash::FileManager::appendParameterFile( "general.param" );
  Ash::FileManager::appendParameterFile( "adrsolver.param" );

  // load the session parameters
  const std::string session = Dune::Parameter::getValue< std::string >( "ash.session" );
  if( Dune::fileExists( session + "/adr.param" ) )
    Dune::Parameter::append( session + "/adr.param" );

  // get number of desired threads (default is 1)
  Dune::Fem::ThreadManager::setMaxNumberThreads( 1 );

  // ProblemType is a Dune::Function that evaluates to $u_0$ and also has a
  // method that gives you the exact solution.
  typedef DiscretizationTraits::GridType GridType;
  typedef Dune::U0< GridType > ProblemType;
  ProblemType problem;

  // read grid
  const std::string gridfile = Ash::FileManager::macroGridFile();
  Dune::GridPtr< GridType > gridptr( gridfile );
  Dune::Parameter::appendDGF( gridfile );
  GridType &grid = *gridptr;

  // refine the grid until the startLevel is reached
  const int refinements = Dune::Parameter::getValue< int >( "ash.refinements", 0 );
  for( int level = 0; level < refinements; ++level )
    Dune::GlobalRefine::apply( grid, 1 );

  typedef Stepper< GridType, ProblemType > Algorithm;
  Algorithm algorithm( grid, problem );
  Algorithm::DiscreteSpaceType &space = algorithm.space();
  Algorithm::DiscreteFunctionType u( "solution", space );
  algorithm( u ); 
  algorithm.finalize(u);

  // log all parameters
  Dune::Parameter::write( "parameter.log" );

  return 0;
}
catch( const Dune::Exception &e )
{                           
  std::cerr << e << std::endl;
  return 1;
}
catch( ... )
{
  std::cerr << "Generic exception caught." << std::endl;
  return 2;
}                                                    
