/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

#ifndef FEMHOWTO_BASEEVOL_HH
#define FEMHOWTO_BASEEVOL_HH

#include <sstream>

#include <dune/fem/writerutilities.hh>


template< class Traits >
struct AlgorithmBase
{
  typedef typename Traits::GridType GridType;
  typedef typename Traits::GridPartType GridPartType;
  typedef typename Traits::DiscreteSpaceType DiscreteSpaceType;
  typedef typename Traits::DiscreteFunctionType DiscreteFunctionType;
  typedef Dune::GridTimeProvider< GridType > TimeProviderType;

  AlgorithmBase ( GridType &grid )
  : grid_( grid ),
    gridPart_( grid_ ),
    space_( gridPart_ ),
    saveStep_( Dune::Parameter::getValue< double > ( "ash.savestep" ) ),
    saveTime_( 0 )
  {}

  bool willWrite ( TimeProviderType &tp ) const
  {
    const bool truelyWrite = (saveTime_ < tp.time());
    saveTime_ += (truelyWrite ? saveStep_ : 0);
    return truelyWrite;
  }
  
  DiscreteSpaceType &space () { return space_; }

  virtual void initializeStep ( TimeProviderType &tp, DiscreteFunctionType &solution ) = 0;

  virtual void step ( TimeProviderType &tp, DiscreteFunctionType &solution ) = 0;

  virtual void finalizeStep ( TimeProviderType &tp, DiscreteFunctionType &solution ) = 0;

  virtual void operator () ( DiscreteFunctionType &solution )
  {
    const std::string basefilename = "adr.solution";

    const double startTime = Dune::Parameter::getValue< double >( "ash.startTime", 0.0 );
    saveTime_ = startTime;
    const double endTime = Dune::Parameter::getValue<double>( "ash.endTime", 1.0 );
    const std::string session = Dune::Parameter::getValue< std::string >( "ash.session" );

    const std::string path = session + "/adrdata";
    Dune::createDirectory( path );

    TimeProviderType tp( startTime, grid_ ); 
    initializeStep( tp, solution );

    int writeCounter = 0;
    writeSerialXdr< DiscreteFunctionType >( path, basefilename, solution, writeCounter );
    ++writeCounter;

    tp.provideTimeStepUpperBound( 0.5 * saveStep_ );

    for( tp.init(); tp.time() < endTime; tp.next() )
    {
      tp.provideTimeStepUpperBound( 0.5 * saveStep_ );

      step( tp, solution );

      if( !solution.dofsValid() )
        abort();

      if( willWrite( tp ) )
      {
        writeSerialXdr< DiscreteFunctionType >( path, basefilename, solution, writeCounter );
        ++writeCounter;
      }
    } 

    // write last time step  
    writeSerialXdr< DiscreteFunctionType >( path, basefilename, solution, writeCounter );

    // finalize time step 
    finalizeStep( tp, solution );

    // wait for all procs to reach this point 
    grid_.comm().barrier();
    if( grid_.comm().rank() == 0 ) 
    {
      std::string filename =  session + "/adr.done";
      std::ofstream endtoken( filename.c_str() );
      if( endtoken.is_open() ) 
        endtoken << "The eagle has landed on his beagle!" << std::endl;
      else 
        std::cerr << "Couldn't open file '" << filename << "'" << std::endl;
    }
  }

  void finalize ( DiscreteFunctionType &solution ) {}

protected:      
  GridType &grid_;
  GridPartType gridPart_;
  DiscreteSpaceType space_;
  const double saveStep_;
  mutable double saveTime_;
}; 

#endif // #ifndef FEMHOWTO_BASEEVOL_HH
