/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

#ifndef DISCRETEMODELS_HH
#define DISCRETEMODELS_HH

#include <dune/fem/function/adaptivefunction.hh>
#include <dune/fem/gridpart/common/gridpart.hh>
#include <dune/fem/gridpart/adaptiveleafgridpart.hh>
#include <dune/fem/pass/dgdiscretemodel.hh>
#include <dune/fem/pass/dgpass.hh>
#include <dune/fem/pass/selection.hh>
#include <dune/fem/quadrature/cachingquadrature.hh>
#include <dune/fem/solver/timeprovider.hh>
#include <dune/fem/space/dgspace.hh>

#include "../common/discretizationtraits.hh"

namespace Dune
{

  // Internal Forward Declarations
  // -----------------------------

  template< class Model, int polOrd, int passVId, int passUId >
  class AdvDiffDModel1;

  template< class Model, class NumFlux, int polOrd, int passVId, int passUId, int passGradId >
  class AdvDiffDModel2; 


  // General Traits class
  template <class Model,int dimRange,int polOrd>
  class PassTraits
  {
  public:
    typedef typename Model :: Traits                                 ModelTraits;
    typedef typename ModelTraits :: GridPartType                     GridPartType;
    typedef typename GridPartType :: GridType                        GridType;
    static const int dimDomain = Model :: Traits :: dimDomain;

    typedef CachingQuadrature< GridPartType, 0 >                     VolumeQuadratureType;
    typedef CachingQuadrature< GridPartType, 1 >                     FaceQuadratureType;

    typedef typename DiscretizationTraits::SolutionSpaceType::
      ToNewDimRange< dimRange >::Type DiscreteFunctionSpaceType;

    typedef typename DiscreteFunctionSpaceType::FunctionSpaceType FunctionSpaceType;

    typedef AdaptiveDiscreteFunction< DiscreteFunctionSpaceType >    DestinationType;
  };


  // AdvDiffTraits1
  // --------------

  template< class Model, int polOrd, int passVId, int passUId >
  struct AdvDiffTraits1
  {
    typedef typename Model :: Traits                                 ModelTraits;
    typedef typename ModelTraits :: GridType                         GridType;

    enum { dimRange = ModelTraits::dimGradRange };
    enum { dimDomain = ModelTraits::dimDomain };

    typedef PassTraits< Model, dimRange, polOrd >                    Traits;
    typedef typename Traits :: FunctionSpaceType                     FunctionSpaceType;

    typedef typename Traits :: VolumeQuadratureType                  VolumeQuadratureType;
    typedef typename Traits :: FaceQuadratureType                    FaceQuadratureType;
    typedef typename Traits :: GridPartType                          GridPartType;

    typedef typename Traits :: DiscreteFunctionSpaceType             DiscreteFunctionSpaceType;
    typedef typename Traits :: DestinationType                       DestinationType;
    typedef DestinationType                                          DiscreteFunctionType;

    typedef typename DestinationType :: DomainType                   DomainType;
    typedef typename ModelTraits :: FaceLocalCoordinate              FaceLocalCoordinate;
    typedef typename ModelTraits :: LocalCoordinate                  LocalCoordinate;
    typedef typename DestinationType :: RangeType                    RangeType;
    typedef typename DestinationType :: JacobianRangeType            JacobianRangeType;

    typedef AdvDiffDModel1< Model, polOrd, passVId, passUId > DGDiscreteModelType;
  };

  template <class Model, class NumFlux, int polOrd, int passVId, int passUId, int passGradId >
  struct AdvDiffTraits2
  {
    typedef typename Model :: Traits                                 ModelTraits;
    typedef typename ModelTraits :: GridType                         GridType;

    enum { dimRange = ModelTraits::dimRange };
    enum { dimDomain = ModelTraits::dimDomain };

    typedef PassTraits< Model, dimRange, polOrd >                    Traits;
    typedef typename Traits :: FunctionSpaceType                     FunctionSpaceType;

    typedef typename Traits :: VolumeQuadratureType                  VolumeQuadratureType;
    typedef typename Traits :: FaceQuadratureType                    FaceQuadratureType;
    typedef typename Traits :: GridPartType                          GridPartType;
    typedef typename Traits :: DiscreteFunctionSpaceType             DiscreteFunctionSpaceType;
    typedef typename Traits :: DestinationType                       DestinationType;
    typedef DestinationType                                          DiscreteFunctionType;

    typedef typename DestinationType :: DomainType                   DomainType;
    typedef typename ModelTraits :: FaceLocalCoordinate              FaceLocalCoordinate;
    typedef typename ModelTraits :: LocalCoordinate                  LocalCoordinate;
    typedef typename DestinationType :: RangeType                    RangeType;
    typedef typename DestinationType :: JacobianRangeType            JacobianRangeType;

    typedef AdvDiffDModel2< Model, NumFlux, polOrd, passVId, passUId, passGradId > DGDiscreteModelType;
  };



  // AdvDiffDModel1
  // --------------

  template< class Model, int polOrd, int passVId, int passUId >
  class AdvDiffDModel1
  : public DGDiscreteModelDefaultWithInsideOutside< AdvDiffTraits1< Model, polOrd, passVId, passUId >, passVId, passUId >
  {
    typedef DGDiscreteModelDefaultWithInsideOutside< AdvDiffTraits1< Model, polOrd, passVId, passUId >, passVId, passUId > BaseType;

    using BaseType::numericalFlux;
    using BaseType::boundaryFlux;

    // This type definition allows a convenient access to arguments of passes.
    integral_constant< int, passUId > uVar;

  public:
    typedef AdvDiffTraits1< Model, polOrd, passVId, passUId > Traits;

    typedef typename Traits :: DomainType                            DomainType;
    typedef typename Traits :: LocalCoordinate                       LocalCoordinate;
    typedef typename Traits :: FaceLocalCoordinate                   FaceLocalCoordinate;
    typedef typename Traits :: RangeType                             RangeType;
    typedef typename Traits :: GridType                              GridType;
    typedef typename Traits :: JacobianRangeType                     JacobianRangeType;
    typedef typename Traits :: GridPartType :: IntersectionType      Intersection;
    typedef typename Traits :: GridPartType
              :: IntersectionIteratorType                            IntersectionIterator;
    typedef typename GridType :: template Codim< 0 > :: Entity       EntityType;

  public:
    explicit AdvDiffDModel1 ( const Model &model )
    : model_( model ),
      cflDiffinv_( 2 * (polOrd + 1) )
    {}

    bool hasSource () const { return false; }
    bool hasFlux () const { return true; } 

    template< class ArgumentTuple >
    double numericalFlux ( const Intersection &it, double time, const FaceLocalCoordinate &x,
                           const ArgumentTuple &uLeft, const ArgumentTuple &uRight,
                           RangeType &gLeft, RangeType &gRight )
    {
      JacobianRangeType diffmatrix;
      const DomainType normal = it.integrationOuterNormal( x );

      gLeft = 0;
      model_.diffusion1( this->inside(), time, it.geometryInInside().global( x ), uLeft[ uVar ], diffmatrix );
      diffmatrix.umv( normal, gLeft );
      model_.diffusion1( this->outside(), time, it.geometryInOutside().global( x ), uRight[ uVar ], diffmatrix );
      diffmatrix.umv( normal, gLeft );
      gLeft *= 0.5;
      
      gRight = gLeft;

      // upper bound for the next time step length
      return model_.diffusionTimeStep() * cflDiffinv_;
    }

    template< class ArgumentTuple, class FaceDom >
    double boundaryFlux ( const Intersection &it, double time, const FaceDom &x,
                          const ArgumentTuple &uLeft, RangeType &gLeft )
    {
      JacobianRangeType diffmatrix;
      const DomainType normal = it.integrationOuterNormal( x );

      gLeft = 0;
      model_.diffusion1 ( this->inside(), time, it.geometryInInside().global( x ), uLeft[ uVar ], diffmatrix );
      diffmatrix.umv( normal, gLeft );

      return 0.5 * model_.diffusionTimeStep() * cflDiffinv_;
    }

    template< class ArgumentTuple >
    void analyticalFlux ( const EntityType &en, double time, const LocalCoordinate &x,
                          const ArgumentTuple &u, JacobianRangeType &f )
    {
      model_.diffusion1( en, time, x, u[ uVar ], f );
    } 

  private:
    const Model &model_;
    const double cflDiffinv_;
  };



  template <class Model, class NumFlux, int polOrd, int passVId, int passUId, int passGradId>
  class AdvDiffDModel2 :
    public DGDiscreteModelDefaultWithInsideOutside
      <AdvDiffTraits2<Model, NumFlux, polOrd, passVId, passUId, passGradId >, passVId,
            passUId , passGradId>
  {
    typedef DGDiscreteModelDefaultWithInsideOutside
              < AdvDiffTraits2< Model, NumFlux, polOrd, passVId, passUId, passGradId >, passVId,
                passUId, passGradId >                        BaseType;

    using BaseType :: numericalFlux;
    using BaseType :: boundaryFlux;

    // These type definitions allow a convenient access to arguments of pass.
    integral_constant<int,passVId> veloVar;
    integral_constant<int,passUId> uVar;
    integral_constant<int,passGradId> sigmaVar;
  public:
    typedef AdvDiffTraits2< Model, NumFlux, polOrd, passVId, passUId, passGradId > Traits;

    typedef typename Traits :: DomainType                            DomainType;
    typedef typename Traits :: LocalCoordinate                       LocalCoordinate;
    typedef typename Traits :: FaceLocalCoordinate                   FaceLocalCoordinate;

    typedef typename Traits :: RangeType                             RangeType;
    typedef typename Traits :: GridType                              GridType;
    typedef typename Traits :: JacobianRangeType                     JacobianRangeType;

    typedef typename Traits :: GridPartType :: IntersectionType      Intersection;
    typedef typename Traits :: GridPartType
              :: IntersectionIteratorType                            IntersectionIterator;
    typedef typename GridType :: template Codim< 0 > :: Entity       EntityType;

  public:
    AdvDiffDModel2(const Model& mod,const NumFlux& numf) 
    : model_( mod ),
      numFlux_( numf ),
      penalty_( 2.5 ), 
      cflDiffinv_(2.*(polOrd+1.) )
    {}

    bool hasSource() const { return true; }
    bool hasFlux() const { return true; }

    template <class ArgumentTuple, class JacobianTuple>
    double source( const EntityType& en, 
                 const double time, 
                 const DomainType& x,
                 const ArgumentTuple& u, 
                 const JacobianTuple& jac, 
                 RangeType& s ) const 
    {
			model_.source( en, time, x, jac[veloVar], u[uVar], s );
			return 0.;
    }

    template <class ArgumentTuple>
    double numericalFlux(const Intersection& it,
                         double time, const FaceLocalCoordinate& x,
                         const ArgumentTuple& uLeft,
                         const ArgumentTuple& uRight,
                         RangeType& gLeft,
                         RangeType& gRight)
    {
      const DomainType normal = it.integrationOuterNormal(x);


      double wave = numFlux_.   
        numericalFlux(it, time, x, uLeft[veloVar], uRight[veloVar], uLeft[uVar], uRight[uVar], gLeft, gRight);

      JacobianRangeType diffmatrix;
      RangeType diffflux(0.);
      model_.      
        diffusion2(this->inside(), time, it.geometryInInside().global(x),
                   uLeft[uVar], uLeft[sigmaVar], diffmatrix);
      diffmatrix.umv(normal, diffflux);
      model_.
        diffusion2(this->outside(), time, it.geometryInOutside().global(x),
                   uRight[uVar], uRight[sigmaVar], diffmatrix);
      diffmatrix.umv(normal, diffflux);
      diffflux*=0.5;

      double factor = penalty_ * model_.diffusionCoefficient() *
                      normal.two_norm2() /
                      ( 0.5 * ( this->enVolume()+this->nbVolume() ) );

      RangeType jump( uLeft[uVar] );
      jump -= uRight[uVar];
      diffflux.axpy(factor,jump);

      gLeft  += diffflux;
      gRight += diffflux;
      return std::max( wave, factor*cflDiffinv_ );
    }

    template <class ArgumentTuple>
    double boundaryFlux(const Intersection& it,
                        const double time, 
                        const FaceLocalCoordinate& x,
                        const ArgumentTuple& uLeft,
                        RangeType& gLeft)
    {
      const DomainType normal = it.integrationOuterNormal(x);
      typedef typename ArgumentTuple::template Get<passUId>::Type UType;
      double wave=0.0;

      model_.boundaryFlux2( it, time, x, uLeft[veloVar], uLeft[uVar], gLeft);
      gLeft *= normal.two_norm();

      return wave;
    }

    template <class ArgumentTuple> 
    void analyticalFlux( const EntityType& en,
                         const double time, 
                         const LocalCoordinate& x,
                         const ArgumentTuple& u, JacobianRangeType& f )
    {
      model_.advection(en,time, x, u[veloVar], u[uVar],f);
      JacobianRangeType diffmatrix;
      model_.diffusion2(en, time, x, u[uVar], u[sigmaVar], diffmatrix);
      f += diffmatrix;
    }
  private:
    const Model   &model_;
    const NumFlux &numFlux_;
    const double   penalty_;
    const double   cflDiffinv_;
  };

} // namespace Dune

#endif // #ifndef DISCRETEMODELS_HH
