/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

#ifndef PROBLEM_ASH_HH
#define PROBLEM_ASH_HH

#include <dune/common/fmatrix.hh>

#include <dune/fem/space/common/functionspace.hh>

namespace Dune
{

  template< class GridType >
  struct U0 
  : public Fem::Function< Dune::FunctionSpace< double, double, GridType::dimensionworld, 1 >, U0< GridType > >
  { 
    typedef U0< GridType > ThisType;
    typedef Fem::Function< Dune::FunctionSpace< double, double, GridType::dimensionworld, 1 >, U0< GridType > > BaseType;

  public:
    //static const int dimDomain = GridType::dimensionworld;
    //static const int dimRange = 1;

    typedef typename BaseType::FunctionSpaceType                  FunctionSpaceType; 
    typedef typename FunctionSpaceType::DomainType                DomainType;
    typedef typename FunctionSpaceType::RangeType                 RangeType;
    typedef typename FunctionSpaceType::DomainFieldType           DomainFieldType;
    typedef typename FunctionSpaceType::RangeFieldType            RangeFieldType;
    typedef typename GridType :: template Codim< 0 > :: Entity      Entity;

    U0 () 
    : startTime_( Parameter::getValue< double >( "ash.startTime",0.0 ) ),
      concentration_( Parameter::getValue< double >( "ash.volcano.concentration" ) ),
      radius_( Parameter::getValue< double >( "ash.volcano.radius" ) ),
      diffusion_( Parameter::getValue< double >( "ash.diffusion", 0.01 ) )
    {                                                            
      position_ = Parameter::getValue< DomainType >( "ash.volcano.position" );
    }

    double startTime () const { return startTime_; }

    double diffusion () const { return diffusion_; }

    //void velocity ( const DomainType &x, DomainType &v ) const { v = 0; }

    void evaluate ( const DomainType &x, RangeType &res ) const
    {
      evaluate( x, startTime_, res );
    }

    void evaluate ( const DomainType &x, double t, RangeType &res ) const
    {
      res = 0;
    }

    void evaluateRhs ( const DomainType &x, double time, RangeType &source ) const
    {
      source = 0;
      DomainType distDisturb ( x );
      distDisturb.axpy( -1, position_ );
      if( distDisturb.two_norm() <= radius_ )
        source += concentration_;

      // scale such that source term is zero at initial time and increases quickly
      source *= (1 - std::exp( -100*time ) );

      // scale additionally with weighted Gaussian so that volcano eruption converges to zero for longer times
      source *= std::exp( -12.5*(time*time) );
    }

  private:
    double         startTime_;
    DomainType     position_;
    const double   concentration_;
    const double   radius_;
    const double   diffusion_;
  };

} // namespace Dune

#endif // #ifndef PROBLEM_ASH_HH
