/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

#ifndef STEPPER_HH
#define STEPPER_HH

#include <config.h>

#include <iostream>
#include <string>

#include <dune/common/nullptr.hh>

#include <dune/fem/misc/l2error.hh>
#include <dune/fem/operator/projection/l2projection.hh>
#include <dune/fem/solver/odesolver.hh>

#include "../common/discretizationtraits.hh"
#include "baseevolution.hh"
#include "models.hh"

// approximation order
const int order   = POLORDER;
const int rkSteps = POLORDER + 1;

template< class GridImp, class InitialDataType >
struct StepperTraits
{
  // type of Grid
  typedef GridImp GridType;
  typedef DiscretizationTraits::GridPartType GridPartType;

  // An analytical version of our model
  typedef AdvectionDiffusionModel< GridPartType, InitialDataType >   ModelType;

  // The flux for the discretization of advection terms
  typedef UpwindFlux< ModelType >                                    FluxType;

  // The DG Operator (using 2 Passes)
  typedef Dune::DGAdvectionDiffusionOperator< ModelType, UpwindFlux,
                                              order >                DgType; 

  // The discrete function for the unknown solution is defined in the DgOperator
  typedef typename DgType::DestinationType                         DiscreteFunctionType;
  // ... as well as the Space type
  typedef typename DgType::SpaceType                               DiscreteSpaceType;

  // The ODE Solvers                                                         
  typedef DuneODE::OdeSolverInterface< DiscreteFunctionType >      OdeSolverInterfaceType;
  typedef DuneODE::ImplicitOdeSolver< DiscreteFunctionType >       ImplicitOdeSolverType;

  // type of restriction/prolongation projection for adaptive simulations 
  typedef Dune::RestrictProlongDefault< DiscreteFunctionType > RestrictionProlongationType;
};

template< class GridImp, class InitialDataType >
struct Stepper
: public AlgorithmBase< StepperTraits< GridImp, InitialDataType> >
{
  // my traits class 
  typedef StepperTraits< GridImp, InitialDataType> Traits ;

  // my base class 
  typedef AlgorithmBase < Traits > BaseType;

  // type of Grid
  typedef typename Traits::GridType                  GridType;

  // An analytical version of our model
  typedef typename Traits::ModelType                 ModelType;

  // The flux for the discretization of advection terms
  typedef typename Traits::FluxType                  FluxType;

  // The DG Operator (using 2 Passes)
  typedef typename Traits::DgType                    DgType;

  // The discrete function for the unknown solution is defined in the DgOperator
  typedef typename Traits::DiscreteFunctionType      DiscreteFunctionType;

  // ... as well as the Space type
  typedef typename Traits::DiscreteSpaceType         DiscreteSpaceType;

  // The ODE Solvers
  typedef typename Traits::OdeSolverInterfaceType    OdeSolverInterfaceType;
  typedef typename Traits::ImplicitOdeSolverType     ImplicitOdeSolverType;

  typedef typename BaseType::TimeProviderType   TimeProviderType;

  using BaseType::grid_;

  // brief constructor of the stepper class
  Stepper(GridType& grid, InitialDataType& problem) :
    BaseType ( grid ),
    problem_(problem),
    model_(problem_),
    convectionFlux_(model_),
    dgOperator_(grid_, convectionFlux_),
    odeSolver_(0)
  {}

  ~Stepper () { delete odeSolver_; }

  InitialDataType &problem () const { return *problem_; }

  const ModelType &model () const { return model_; }

  // before first step, do data initialization 
  void initializeStep(TimeProviderType& tp, DiscreteFunctionType& u)
  {
    // create implicit or explicit ode solver
    odeSolver_ = new ImplicitOdeSolverType(dgOperator_, tp, rkSteps);
    assert( odeSolver_ );

    Dune::Fem::L2Projection< InitialDataType, DiscreteFunctionType > l2pro;
    l2pro(problem_, u);                                  

    // odeSolver_->initialize applies the DG Operator once to get an initial
    // estimate on the time step. This does not change the initial data u.
    odeSolver_->initialize(u);                          
  }

  // solve ODE for one time step 
  void step(TimeProviderType& tp, DiscreteFunctionType& u) 
  {
    assert(odeSolver_);
    odeSolver_->solve(u);                              
  }

  // after last step, do EOC calculation 
  void finalizeStep(TimeProviderType& tp, DiscreteFunctionType& u)
  {
    // delete ode solver
    delete odeSolver_;
    odeSolver_ = 0;
  }

private:
  // InitialDataType is a Dune::Operator that evaluates to $u_0$ and also has a
  // method that gives you the exact solution.
  InitialDataType  &problem_;
  ModelType               model_;
  // Initial flux for advection discretization (UpwindFlux)
  FluxType                convectionFlux_;
  DgType                  dgOperator_;
  OdeSolverInterfaceType *odeSolver_;
};



#endif // STEPPER_HH
