/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

// vim: set expandtab ts=2 sw=2 sts=2:
#ifndef FRONTENDMODE_HH
#define FRONTENDMODE_HH

#include <cassert>
#include <iostream>
#include <map>
#include <string>

#include <SDL.h>

#include <dune/ash/opengl/texture.hh>
#include <dune/ash/widget/widget.hh>

#include "sharedinfo.hh"

class FrontendMode;



// FrontEndModeStack
// -----------------

class FrontendModeStack
{
  typedef FrontendModeStack ThisType;

  typedef std::vector< FrontendMode * > StackType;

public:
  typedef Dune::Ash::Widget::Ticks Ticks;
  typedef std::map< std::string, std::string > InfoMap;

  enum ExitStatus { StatusRestart = 0, StatusShutdown = 1, StatusError = 0x10 };

  FrontendModeStack ( SDL_Surface *screen, SharedInformation *sharedInformation );

  ~FrontendModeStack () 
  { 
    clear(); 
    collectGarbage();
  }

  bool empty () const { return stack_.empty(); }

  const FrontendMode &top () const { return *stack_.back(); }
  FrontendMode &top () { return *stack_.back(); }

  void clear ();
  void push ( FrontendMode *mode );
  void pop ();

  void tick ( Ticks ticks );

  void setLanguage ( const std::string &language );
  std::string getLanguage () const { return language_ ; }

  FrontendMode *father ( FrontendMode *mode );

  SDL_Surface *screen () const { return screen_; }
  
  SharedInformation *sharedInformation () { return sharedInformation_; } 

  const Dune::Ash::Texture &mapTexture () const { return mapTexture_; }
  const Dune::Ash::Texture &equationTexture () const { return equationTexture_; }
  const Dune::Ash::Texture &flagsTexture () const { return flagsTexture_; }

  Dune::Ash::BillBoard icon ( int icon );
  Dune::Ash::BillBoard icon3x2 ( int icon );

  Dune::Ash::BillBoard flag ( int flag );

  void setExitStatus ( ExitStatus status ) { exitStatus_ = status; }  
  const ExitStatus exitStatus () const { return exitStatus_; }

  const InfoMap &infoMap () const { return infoMap_; }

private:
  void collectGarbage ();

  FrontendModeStack ( const ThisType & );
  const ThisType &operator= ( const ThisType & );

  SDL_Surface *screen_;
  StackType stack_;
  std::stack< FrontendMode * > garbage_;

  SharedInformation *sharedInformation_;

  std::string language_;

  Dune::Ash::Texture mapTexture_;
  Dune::Ash::Texture equationTexture_;
  Dune::Ash::Texture flagsTexture_;
  Dune::Ash::Texture iconsTexture_;

  InfoMap infoMap_;

  ExitStatus exitStatus_;
};



/*
 * This file defines a basic virtual interface for a mode of the userinterface.
 * Modes will be - in our case - "drawing mode" (the user can draw the vector field
 * with the mouse) and "visualization mode" (the userinterface draws the calculations
 * which are performed by the backend)
 *
 * This interface should describe anything which is needed to communicate from the low level
 * class MainWindow (which does most of the SDL stuff) to the more abstract layers...
 */
class FrontendMode
: public Dune::Ash::Widget
{
  typedef FrontendMode ThisType;

public:
  explicit FrontendMode ( FrontendModeStack &modeStack );
  virtual ~FrontendMode (); 

  // this is invoked to set up the mode (i.e. only once, not before each frame)
  virtual void initializeMode () {}

  // cleanUp is invoked when the mode is popped
  // note: The mode is deleted later by the garbage collector.
  virtual void cleanUp () {}

  // ticking
  virtual void tick ( Ticks ticks );

  virtual void setLanguage( const std::string &language );

  /*
   * Some simple accessor methods
   */
  const FrontendModeStack &modeStack () const { return modeStack_; }
  FrontendModeStack &modeStack () { return modeStack_; }

  SDL_Surface *mainScreen () const { return modeStack().screen(); }

  virtual void loadSession( const int & sessionID ) {}

  int screenWidth () const { return mainScreen()->w; }
  int screenHeight () const { return mainScreen()->h; }

private:
  // prohibit copying and assignment
  FrontendMode ( const ThisType & );
  const ThisType &operator= ( const ThisType & );

  // data fields
  FrontendModeStack &modeStack_;
};



// Implementation of FrontendModeStack
// -----------------------------------

inline void FrontendModeStack::push ( FrontendMode *mode )
{
  assert( mode );
  stack_.push_back ( mode );
  top().initializeMode();
}


inline void FrontendModeStack::pop ()
{
  stack_.back()->cleanUp();
  garbage_.push( stack_.back() );
  stack_.pop_back();
  if( !empty() )
    top().initializeMode();
}


inline Dune::Ash::BillBoard FrontendModeStack::icon ( int icon )
{
  Dune::FieldVector< GLfloat, 2 > texUpperLeft, texLowerRight;
  texUpperLeft[ 0 ] = GLfloat( icon % 4 )*0.25f;
  texUpperLeft[ 1 ] = GLfloat( icon / 4 )*0.25f;
  texLowerRight[ 0 ] = texUpperLeft[ 0 ] + 0.25f;
  texLowerRight[ 1 ] = texUpperLeft[ 1 ] + 0.25f;

  return Dune::Ash::BillBoard( iconsTexture_, texUpperLeft, texLowerRight );
}

inline Dune::Ash::BillBoard FrontendModeStack::icon3x2 ( int icon )
{
  Dune::FieldVector< GLfloat, 2 > texUpperLeft, texLowerRight;
  texUpperLeft[ 0 ] = GLfloat( icon % 4 )*0.25f;
  texUpperLeft[ 1 ] = GLfloat( icon / 4 )*0.25f;
  texLowerRight[ 0 ] = texUpperLeft[ 0 ] + 0.25f;
  texLowerRight[ 1 ] = texUpperLeft[ 1 ] + 0.25f;

  return Dune::Ash::BillBoard( Dune::Ash::makeFieldVector( 1. / 6., 0., 0. ), Dune::Ash::makeFieldVector( 5. / 6., 1., 0. ), iconsTexture_, texUpperLeft, texLowerRight );
}

inline Dune::Ash::BillBoard FrontendModeStack::flag ( int flag )
{
  Dune::FieldVector< GLfloat, 2 > texUpperLeft, texLowerRight;
  texUpperLeft[ 0 ] = GLfloat( flag % 4 )*0.25f;
  texUpperLeft[ 1 ] = GLfloat( flag / 4 )*0.25f;
  texLowerRight[ 0 ] = texUpperLeft[ 0 ] + 0.25f;
  texLowerRight[ 1 ] = texUpperLeft[ 1 ] + 0.25f;

  return Dune::Ash::BillBoard( flagsTexture_, texUpperLeft, texLowerRight );
}

#endif // #ifndef FRONTENDMODE_HH
