/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

// vim: set expandtab ts=2 sw=2 sts=2: 
#ifndef GRIDVISUALIZATION_HH
#define GRIDVISUALIZATION_HH

#include <cassert>

#include <dune/ash/opengl/opengl.hh>

namespace Dune
{

  namespace Ash
  {

    struct GridVisualization
    {
      /*
       * Draw the grid. For an explanation of barycentricWarpFactor see void drawCell ( ... )
       */
      template< class GridPart >
      static void drawGrid ( const GridPart &gridPart, double barycentricWarpFactor = 1.0 )
      {
        typedef typename GridPart::template Codim< 0 >::IteratorType Iterator;
        const Iterator end = gridPart.template end< 0 >();
        for( Iterator it = gridPart.template begin< 0 >(); it != end; ++it )
          drawElement( it->geometry(), barycentricWarpFactor );
      }

      /*
       * Draw a cell. Each corner is "warped" with respect to the barycenter of the cell. Factors
       * smaller than 1. make the drawn cell smaller...
       */
      template< class Geometry >
      static void drawElement ( const Geometry &geometry, double barycentricWarpFactor = 1.0 )
      {
        const int corners = geometry.corners();
        assert( (corners == 3) || (corners == 4) );

        typename Geometry::GlobalCoordinate corner[ 4 ];
        corner[ 0 ] = geometry.corner( 0 );
        corner[ 1 ] = geometry.corner( 1 );
        corner[ 2 ] = geometry.corner( 2 );
        corner[ 3 ] = geometry.corner( 3 );

        // warp the corner coordinates towards the middle
        typename Geometry::GlobalCoordinate center( corner[ 0 ] );
        for( int i = 1; i < corners; ++i )
          center += corner[ i ];
        center *= 0.25;
        for( int i = 0; i < corners; ++i )
          corner[ i ].axpy( 1. - barycentricWarpFactor, center - corner[ i ] );

        glLineWidth( 1. );
        glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );

        glBegin( corners == 4 ? GL_QUADS : GL_TRIANGLES );
        {
          /* 
           * We have to respect the ordering of corners - a simple for loop (i.e. corners 0,1,2,3)
           * in conjunction with GL_LINE_LOOP would not result in a square being drawn.
           */
          glVertex( corner[ 0 ] );
          glVertex( corner[ 1 ] );
          if( corners == 4 )
            glVertex( corner[ 3 ] );
          glVertex( corner[ 2 ] );
        }
        glEnd();

        glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
      }
    };

  } // namespace Ash

} // namespace Dune

#endif // #ifndef GRIDVISUALIZATION_HH
