/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

// vim: set expandtab ts=2 sw=2 sts=2:
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <dune/common/nullptr.hh>

#include <dune/fem/io/parameter.hh>

#include <dune/ash/opengl/displaylist.hh>
#include <dune/ash/fvector.hh>
#include <dune/ash/widget/button.hh>

#include "languagemode.hh"

using Dune::Ash::Rectangle;
using Dune::Ash::makeFieldVector;
using Dune::Ash::ButtonWidget;

static std::vector< bool > drawLine ()
{
  std::vector< bool > drawLine;
  drawLine.push_back( false );
  drawLine.push_back( false );
  drawLine.push_back( false );
  drawLine.push_back( false );
  return drawLine;
}

static std::vector< int > splitting ()
{
  std::vector< int > splitting;
  splitting.push_back( 1 );
  splitting.push_back( 2 );
  splitting.push_back( 6 );
  splitting.push_back( 1 );
  
  return splitting;
}

static std::vector< std::string > languangeArray ()
{
  std::vector< std::string > languangeArray;
  languangeArray.push_back( "de" );
  languangeArray.push_back( "en" );
  
  return languangeArray;
}


struct LanguageMode::EventClose
: public Dune::Ash::Event
{
  explicit EventClose ( FrontendModeStack &modeStack, const int lang )
  : modeStack_( modeStack ),
    lang_( lang )
  {}

  void executeEvent () 
  {
    modeStack_.setLanguage( languangeArray()[lang_] ); 
    modeStack_.pop(); 
  }

private:
  FrontendModeStack &modeStack_;
  const int lang_;
};


// Implementation of LanguageMode
// --------------------------
LanguageMode::LanguageMode( FrontendModeStack &modeStack )
: FrontendMode( modeStack ),
  father_( nullptr ),
  shrinkWidget_( 12.0f / 32.0f, 5.0f / 18.0f ),
  text_( 120.0f, makeFieldVector( 6.0f, 0.5f ) ),
  split_( splitting(), drawLine() ),
  languageArray_( 4 , 1, makeFieldVector( 1.0f, 1.0f, 1.0f ), makeFieldVector( 0.08f, 0.2f ) ),
  placeholder_( Rectangle( makeFieldVector( 1.0f, 1.0f, 1.0f, 1.0f ) ) ),
  parser_( "panel" )
{
  setLanguage( modeStack.getLanguage() );

  shrinkWidget_.setChild( &split_ );

  split_.setChild( 0, &placeholder_ );
  split_.setChild( 1, &text_ );
  split_.setChild( 2, &languageArray_ );
  split_.setChild( 3, &placeholder_ );

  languageArray_.setChild( 1, 0, new ButtonWidget( new EventClose( modeStack, 0 ), modeStack.flag( 0 ) ) );
  languageArray_.setChild( 2, 0, new ButtonWidget( new EventClose( modeStack, 1 ), modeStack.flag( 1 ) ) ); 
}

LanguageMode::~LanguageMode ()
{
  for( std::size_t v = 0; v < languageArray_.size()[ 1 ]; ++v )
  {
     for( std::size_t h = 0; h < languageArray_.size()[ 0 ]; ++h )
    {
      Widget *widget = languageArray_.setChild( h, v, nullptr );

      if( !widget )
        continue;

      Dune::Ash::ButtonWidget *button = dynamic_cast< Dune::Ash::ButtonWidget * >( widget );
      if( button )
        delete button->setEvent( nullptr );

      delete widget;
    }
  }
}

void LanguageMode::setLanguage ( const std::string &language )
{
  parser_( language, textMap_ );
  std::istringstream input( textMap_[ "macros"] + textMap_[ "language" ] );
  text_.setText( Ash::FileManager::dataPath() + "/text", input );
}

void LanguageMode::initializeMode ()
{
  father_ = modeStack().father( this );
}

void LanguageMode::draw ()
{
  if( father_ )
    father_->draw();
  shrinkWidget_.draw();
}

void LanguageMode::tick ( Ticks ticks )
{
  if( father_ )
    father_->tick( ticks );
  shrinkWidget_.tick( ticks );
}
