/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

// vim: set expandtab ts=2 sw=2 sts=2:
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <dune/fem/io/io.hh>

#include <dune/ash/image.hh>
#include <dune/ash/fvector.hh>
#include <dune/ash/opengl/opengl.hh>

#include "infomode.hh"
#include "mainmode.hh"
#include "volcanomode.hh"


// MainMode::EventNextMode
// -----------------------

struct MainMode::EventNextMode
: public Dune::Ash::Event
{
  explicit EventNextMode ( MainMode &mode )
  : mode_( mode )
  {}

  void executeEvent ()
  {
    mode_.sharedInformation_->reset();
    mode_.modeStack().push( new VolcanoMode( mode_.modeStack() ) );
  }

private:
  MainMode &mode_;
};



// MainMode::EventLink
// -------------------

struct MainMode::EventLink
: public Dune::Ash::Event
{
  EventLink ( FrontendModeStack &modeStack, const std::string &panel )
  : modeStack_( modeStack ),
    panel_( panel )
  {}

  void executeEvent () 
  {
    modeStack_.push( new InfoMode( modeStack_, panel_ ) );
  }

private:
  FrontendModeStack &modeStack_;
  std::string panel_;
};



// Implementation of MainMode
// --------------------------

MainMode::MainMode ( FrontendModeStack &modeStack )
: FrontendMode( modeStack ),
  sharedInformation_( modeStack.sharedInformation() ),
  splitWidget_( 9, 7 ),
  text_( Dune::Ash::makeFieldVector( 120.0f, 120.0f ), Dune::Ash::makeFieldVector( 9.0f, 9.0f ) ),
  panel_( modeStack, Panel::IdMainMode )
{
  text_.setText( Ash::FileManager::dataPath() + "/text", std::string( "news." ) + modeStack.getLanguage() );
  for( InfoIterator it = infoMap().begin(); it != infoMap().end(); ++it )
    text_.setEvent( std::string( "link:" ) + it->first, new EventLink( modeStack, it->first ) );

  splitWidget_.setLeftChild( &text_ );
  splitWidget_.setRightChild( &panel_ );

  panel_.setButton( 6, new Dune::Ash::ButtonWidget( new EventNextMode( *this ), modeStack.icon3x2( 6 ) ) );
}


MainMode::~MainMode ()
{
  text_.deleteEvents();
}


void MainMode::setLanguage ( const std::string &language ) 
{
  panel_.setLanguage ( language );
  text_.setText( Ash::FileManager::dataPath() + "/text", std::string( "news." ) + language );

  text_.deleteEvents();
  for( InfoIterator it = infoMap().begin(); it != infoMap().end(); ++it )
    text_.setEvent( std::string( "link:" ) + it->first, new EventLink( modeStack(), it->first ) );

  // inject events into panel_ text_ once there is access to it
}


void MainMode::draw () 
{
  splitWidget_.draw();
}
