/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

// vim: set expandtab ts=2 sw=2 sts=2:
#ifndef MAINWINDOW_HH
#define MAINWINDOW_HH

#include <SDL.h>

#include <cassert>
#include <sys/time.h>

#include <dune/ash/cursor.hh>
#include <dune/ash/opengl/opengl.hh>
#include <dune/ash/opengl/displaylist.hh>
#include <dune/ash/opengl/texture.hh>

#include "frontendmode.hh"
#include "sharedinfo.hh"

/*
 * Helper class that frees resources on destruction
 */
class SDLHelper
{
public:

  SDLHelper ()
  : SDLInitialized_( false )
  {}

  bool initializeSDL ()
  {
    if ( SDL_Init( SDL_INIT_VIDEO ) != 0 )
    {
      char *errorMsg = SDL_GetError();
      std::cerr << "Error in SDL_Init: " << errorMsg << std::endl;
      SDL_Quit();

      SDLInitialized_ = false;
      return false;
    }

    SDLInitialized_ = true;
    return true;
  }

  ~SDLHelper ()
  {
    if ( SDLInitialized_ )
      SDL_Quit();
  }

private:
  bool SDLInitialized_;
};




/*
 * This models the main window. Window management and user input are - on a low level -
 * done here.
 */
class MainWindow
{
  typedef Dune::Ash::Widget::Ticks Ticks;
  typedef Dune::Ash::Widget::Position Position;

public:
  MainWindow ();
  ~MainWindow ();

  int width () const { return windowWidth_; }
  int height () const { return windowHeight_; }

  SDL_Surface *screen () const
  {
    assert( screen_ );
    return screen_;
  }

  // used to initialize the window. Returns true if initialization went well, false otherwise
  bool initializeWindow ();

  // this performs the main loop of the GUI program
  int run ();

  const FrontendMode &mode () const { return modeStack().top(); }
  FrontendMode &mode () { return modeStack().top(); }

  const FrontendModeStack &modeStack () const { return *modeStack_; }
  FrontendModeStack &modeStack () { return *modeStack_; }

private:
  // forbidden std methods
  MainWindow ( const MainWindow& );
  MainWindow& operator=( const MainWindow& );

  /*
   * private methods
   */
  Ticks ticksSinceEpoch () const
  {
    timeval tv;
    gettimeofday( &tv, 0 );
    return 1000000*static_cast< Ticks >( tv.tv_sec )
           + static_cast< Ticks >( tv.tv_usec );
  }

  // This method is called when the userinterface quits
  void actionOnQuit () const;

  /*
   * Compiles a new OpenGL display list into list, which does the following:
   * texCoords[0,1] are the upper left and texCoords[2,3] the lower right coordinates of
   * a rectangle in the collage texture. This part is drawn to the screen into the
   * rectangle which is specified by screenCoords (same format as above).
   * screenCoords are the actual pixel coordinates on the window...
   * alpha designates the alpha value of the texture
   */
  void compileDrawOfCollagePart ( GLuint list,
                                  GLfloat alpha,
                                  unsigned int *textureCoords,
                                  unsigned int *screenCoords ) const;

  // compiles several parts of the collage
  void compileLogoList ();
  void compileVersionList ();

  /*
   * private data fields
   */
  int windowWidth_;
  int windowHeight_;
  Dune::Ash::MouseCursor *cursor_;
  FrontendModeStack *modeStack_;
  SDLHelper SDLHelper_;
  SDL_Surface *screen_;
  double fps_; // frames per second
  double timePerFrame_; // 1/fps, in microseconds
  SharedInformation sharedInformation_;
  Dune::Ash::Texture iamLogoTexture_, versionTexture_;
  Dune::Ash::DisplayList *versionList_;
};

#endif // MAINWINDOW_HH
