/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

// vim: set expandtab ts=2 sw=2 sts=2:
#include <config.h>

#include <fstream>
#include <iostream>
#include <sstream>
#include <string>

#include <dune/fem/space/common/restrictprolongfunction.hh>
#include <dune/fem/io/io.hh>
#include <dune/fem/io/parameter.hh>

#include <dune/fem/writerutilities.hh>

#include <dune/ash/textures/arrow.hh>
//#include <dune/ash/textures/linearalphamapper.hh>
#include <dune/ash/textures/cloudcolormapper.hh>

#include "sharedinfo.hh"



// Implementation of SharedInformation
// -----------------------------------

SharedInformation::SharedInformation()
: session( "" ),
  volcanoPosition( Dune::Ash::makeFieldVector( 0.16, 0.12 ) ),
  diffusion( -1. ),
  xdr_vector_field_valid_( false ),
  gridPtr_( Ash::FileManager::macroGridFile() ),
  gridPart_( nullptr ),
  windFieldVector_( nullptr ),
  ashSampleVector_( nullptr ),
  arrowList_( nullptr ),
  maxConcentration_( Dune::Parameter::getValue< double >( "ui.concentration.max", 0.1 ) ),
  windFieldDisplayLevel_( 0 ),
  meshDisplayLevel_( 0 )
{}


SharedInformation::~SharedInformation()
{
  // delete textures
  for( AshTextureVectorType::iterator it = ashTextures_.begin(); it != ashTextures_.end(); ++it )
    delete *it;

  delete windFieldVector_;

  if( ashSampleVector_ )
  {
    for ( size_t i=0; i < ashSampleVector_->size(); ++i )
      delete (*ashSampleVector_)[ i ];
  }

  delete gridPart_;
  delete arrowList_;
}


void SharedInformation::initializeArrowList ()
{
  if ( arrowList_ )
    return;

  arrowList_ = new Dune::Ash::DisplayList( Dune::Ash::Arrow() );
}


void SharedInformation::initializeAshSamples () 
{
  if( ashSampleVector_ )
    return;

  ashSampleVector_ = new AshSampleVectorType();
  ashSampleVector_->clear();
}


void SharedInformation::initializeWindField () 
{
  if( windFieldVector_ )
    return;

  if( xdr_vector_field_valid_ )
  {
    windFieldVector_ = new WindFieldVectorType();

    GridType &grid = *gridPtr_;
    const int maxLevel = grid.maxLevel();

    windFieldVector_->resize( maxLevel + 1 );
    for( int level = 0; level <= maxLevel; ++level )
      windFieldVector_->operator[] ( level ) = WindFieldLevelContext( grid, level );

    // read finest wind field level from file
    typedef Dune::Fem::SerialOutputMap< WindFieldType > OutputMap;
    OutputMap map( *windFieldVector_->back().dFunction );
    map.read( session, "windfield.xdr" );

    // restrict wind field to coarser levels
    Dune::RestrictFunction< Dune::Fem::LagrangeLocalRestrictProlong< GridType, 2 > > restrictFunction;
    for( int level = maxLevel; level > 0; --level )
      restrictFunction( *windFieldVector_->operator[] ( level ).dFunction, *windFieldVector_->operator[]( level-1 ).dFunction );

    // create wind field visualizers
    for( int level = 0; level <= maxLevel; ++level )
      windFieldVector_->operator[] ( level ).visualizer = Dune::make_shared< WindFieldVisualizerType >( *arrowList_, *windFieldVector_->operator[]( level ).dFunction );

    windFieldDisplayLevel_ = 1;
  }
}

std::size_t SharedInformation::nextWindFieldDisplayLevel()
{
  if( windFieldVector_ )
    return (windFieldDisplayLevel_ + 1) % (windFieldVector_->size() + 1);
  else
    return 0;
}

void SharedInformation::initializeGrid ()
{
  if( gridPart_ )
    return;

  const int refinements = Dune::Parameter::getValue< int >( "ash.refinements", 0 );
  GridType &grid = *gridPtr_;
  for( int i = 0; i < refinements; ++i )
    Dune::GlobalRefine::apply( grid, 1 );
  gridPart_ = new GridPartType( grid, grid.maxLevel() );

  meshDisplayLevel_ = 0;
}

std::size_t SharedInformation::nextMeshDisplayLevel()
{
  if( windFieldVector_ )
    return (meshDisplayLevel_ + 1) % (windFieldVector_->size());
  else
    return 0;
}

void SharedInformation::loadAshSampleFromFile ( SolutionVisualizerType *visualizer )
{
  // append it to ashSampleVector_
  ashSampleVector_->push_back( visualizer );

  // generate the texture
  //typedef Dune::Ash::LinearAlphaMapper< double > ColorMapper;
  typedef Dune::Ash::CloudColorMapper< double > ColorMapper;
  const ColorMapper colorMapper( 0.0, maxConcentration_ );
  Dune::Ash::Texture *texture = new Dune::Ash::Texture();
  visualizer->loadTexture( colorMapper, *texture );
  ashTextures_.push_back( texture );
}


void SharedInformation::resetSamples ()
{
  for( AshTextureVectorType::iterator it = ashTextures_.begin(); it != ashTextures_.end(); ++it )
    delete *it;
  ashTextures_.clear();

  if( ashSampleVector_ )
  {
    for ( size_t i=0; i < ashSampleVector_->size(); ++i )
      delete (*ashSampleVector_)[ i ];

    delete ashSampleVector_;
    ashSampleVector_ = nullptr;
  }
}


void SharedInformation::reset ()
{
  // bump all generated info
  volcanoPosition = Dune::Ash::makeFieldVector( 0.16, 0.12 ); 

  delete windFieldVector_;
  windFieldVector_ = nullptr;

  diffusion = -1.;
  windFieldDisplayLevel_ = 0;
  meshDisplayLevel_ = 0;

  for( AshTextureVectorType::iterator it = ashTextures_.begin(); it != ashTextures_.end(); ++it )
    delete *it;
  ashTextures_.clear();

  if( ashSampleVector_ )
  {
    for ( size_t i=0; i < ashSampleVector_->size(); ++i )
      delete (*ashSampleVector_)[ i ];

    delete ashSampleVector_;
    ashSampleVector_ = nullptr;
  }
}
