/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

// vim: set expandtab ts=2 sw=2 sts=2:
#ifndef SHAREDINFO_HH
#define SHAREDINFO_HH

#include <fstream>
#include <iostream>
#include <sstream>
#include <string>

#include <dune/common/exceptions.hh>
#include <dune/common/timer.hh>

#include <dune/grid/common/gridinfo.hh>
#include <dune/grid/spgrid/grid.hh>
#include <dune/grid/io/file/dgfparser/gridptr.hh>
#include <dune/grid/common/gridinfo.hh>

#include <dune/fem/gridpart/levelgridpart.hh>

#include <dune/ash/fvector.hh>
#include <dune/ash/opengl/displaylist.hh>

#include "../common/filemanager.hh"
#include "../common/discretizationtraits.hh"
#include "../windfield/includes.hh"


class SharedInformation
{
public:
  typedef DiscretizationTraits::GridType GridType;
  typedef DiscretizationTraits::GridPartType GridPartType;

  typedef DiscretizationTraits::SolutionVisualizerType SolutionVisualizerType;
  typedef std::vector< SolutionVisualizerType * > AshSampleVectorType;
  typedef std::vector< Dune::Ash::Texture * > AshTextureVectorType;

  typedef DiscretizationTraits::WindFieldSpaceType WindFieldSpaceType;
  typedef DiscretizationTraits::WindFieldType WindFieldType;
  typedef DiscretizationTraits::WindFieldVisualizerType WindFieldVisualizerType;

  struct WindFieldLevelContext
  {
    WindFieldLevelContext ()   
    {}

    WindFieldLevelContext ( GridType &grid, int level )
    : gridPart( new GridPartType( grid, level ) ),
      dfSpace( new WindFieldSpaceType( *gridPart ) ),
      dFunction( new WindFieldType( name( level ), *dfSpace ) )
    {}

    static std::string name ( int level )
    {
      std::ostringstream s;
      s << "level " << level << " wind field";
      return s.str();
    }

    Dune::shared_ptr< GridPartType > gridPart;
    Dune::shared_ptr< WindFieldSpaceType > dfSpace;
    Dune::shared_ptr< WindFieldType > dFunction;
    Dune::shared_ptr< WindFieldVisualizerType > visualizer;
  };

  typedef std::vector< WindFieldLevelContext > WindFieldVectorType;

  // public memeber
  std::string session;
  Dune::FieldVector< double, 2 > volcanoPosition;
  double diffusion;
  bool xdr_vector_field_valid_;

  Dune::GridPtr< GridType > gridPtr_;
  GridPartType *gridPart_;

  WindFieldVectorType *windFieldVector_;

  AshSampleVectorType *ashSampleVector_;
  AshTextureVectorType ashTextures_;

  Dune::Ash::DisplayList *arrowList_;

  explicit SharedInformation ();

  ~SharedInformation();

  std::size_t nextWindFieldDisplayLevel();
  std::size_t windFieldDisplayLevel() const { return windFieldDisplayLevel_; }
  void incrementWindFieldDisplayLevel() { windFieldDisplayLevel_ = nextWindFieldDisplayLevel(); }

  std::size_t nextMeshDisplayLevel();
  std::size_t meshDisplayLevel() const { return meshDisplayLevel_; }
  void incrementMeshDisplayLevel() { meshDisplayLevel_ = nextMeshDisplayLevel(); }


  // write intermediat parameters 
  void writeParameters ( std::ostream &stream ) const
  {
    Dune::FieldVector< double, 2 > vp;
    vp[ 0 ] = volcanoPosition[ 0 ];
    vp[ 1 ] = 1 - volcanoPosition[ 1 ];
    stream << "ash.volcano.position: " << vp << std::endl;
    stream << "ash.diffusion: " << diffusion << std::endl;
  }

  // initialing methods
  // ------------------

  // init arrowList
  void initializeArrowList ();

  // init Grid
  void initializeGrid ();

  // init windfield
  void initializeWindField (); 
  
  // init arrowList
  void initializeAshSamples ();

  // load AshSample from File
  void loadAshSampleFromFile ( SolutionVisualizerType *visualizer );
 
  // reset all samples
  void resetSamples();
  
  // reset computed quantities
  void reset();

private:
  double maxConcentration_;

  std::size_t windFieldDisplayLevel_;
  std::size_t meshDisplayLevel_;
};

#endif // #ifndef SHAREDINFO_HH
