/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

// vim: set expandtab ts=2 sw=2 sts=2:
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <fstream>
#include <string>
#include <iostream>

#include <dune/ash/fvector.hh>
#include <dune/ash/image.hh>

#include "../common/filemanager.hh"
#include "exitmode.hh"
#include "infomode.hh"
#include "languagemode.hh"
#include "titlewidget.hh"



// TitleWidget::EventExit
// ----------------------

struct TitleWidget::EventExit
: public Dune::Ash::Event
{
  explicit EventExit ( FrontendModeStack &modeStack )
  : modeStack_( modeStack )
  {}

  void executeEvent ()
  {
    if( Dune::Parameter::getValue< bool >( "ui.askonexit", false ) )
    {
      modeStack_.push( new ExitMode( modeStack_ ) );
    }
    else
    {
      modeStack_.setExitStatus( FrontendModeStack::StatusRestart );
      modeStack_.clear(); 
    }
  }

private:
  FrontendModeStack &modeStack_;   
};



// TitleWidget::EventLanguage
// --------------------------

struct TitleWidget::EventLanguage
: public Dune::Ash::Event
{
  explicit EventLanguage ( FrontendModeStack &modeStack )
  : modeStack_( modeStack )
  {}

  void executeEvent () { modeStack_.push( new LanguageMode( modeStack_ ) ); }

private:
  FrontendModeStack &modeStack_;   
};



// TitleWidget::EventInfo
// ----------------------

struct TitleWidget::EventInfo
: public Dune::Ash::Event
{
  explicit EventInfo ( FrontendModeStack &modeStack )
  : modeStack_( modeStack )
  {}

  void executeEvent () { modeStack_.push( new InfoMode( modeStack_ ) ); }

private:
  FrontendModeStack &modeStack_;   
};



// Implementation of TitleWidget
// -----------------------------

TitleWidget::TitleWidget ( FrontendModeStack &modeStack ) 
: split_( 15, 6 ),
  buttonArray_( 3, 1, Dune::Ash::GridWidget::Color( 1 ), Dune::Ash::makeFieldVector( 0.05f, 0.1f ) ),
  controlArray_( 0.6666, 0.6666, 0.0f ),
  buttonInfo_( new EventInfo( modeStack ), modeStack.icon( 7 ) ),
  buttonLanguage_( new EventLanguage( modeStack ), modeStack.icon( 10 ) ),
  buttonExit_( new EventExit( modeStack ), modeStack.icon( 1 ) )
{
  Dune::Ash::loadImage( Ash::FileManager::absoluteImageName( "dune_ash.png" ), titleTexture_ );
  title_ = Dune::Ash::BillBoard( 
    Dune::Ash::makeFieldVector( 0.05f, 0.05f + 0.9f *  1.0f / 32.0f, 0.0f ), 
    Dune::Ash::makeFieldVector( 0.95f, 0.05f + 0.9f * 31.0f / 32.0f, 0.0f ), titleTexture_ );

  split_.setLeftChild( &title_ );
  split_.setRightChild( &buttonArray_ );

  controlArray_.setChild( &buttonExit_ );

  if( Dune::Parameter::getValue<bool>( "ui.allowquit", true ) )
  {
    buttonArray_.setChild( 0, 0, &buttonLanguage_ );
    buttonArray_.setChild( 1, 0, &buttonInfo_ );
    buttonArray_.setChild( 2, 0, &controlArray_ );
  }
  else
  {
    buttonArray_.setChild( 1, 0, &buttonLanguage_ );
    buttonArray_.setChild( 2, 0, &buttonInfo_ );
  }
}


TitleWidget::~TitleWidget ()
{
  delete buttonExit_.setEvent( nullptr );
  delete buttonLanguage_.setEvent( nullptr );
  delete buttonInfo_.setEvent( nullptr );
}
