/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

#ifndef ELLIPTIC_HH
#define ELLIPTIC_HH

#include <dune/common/fmatrix.hh>

#include <dune/fem/quadrature/cachingquadrature.hh>
#include <dune/fem/operator/common/operator.hh>

// EllipticOperator
// ----------------

template< class DiscreteFunction, class Model >
struct EllipticOperator
: public Dune::Fem::Operator< DiscreteFunction >
{
  typedef DiscreteFunction DiscreteFunctionType;
  typedef Model ModelType;

protected:
  typedef typename DiscreteFunctionType::DiscreteFunctionSpaceType DiscreteFunctionSpaceType;
  typedef typename DiscreteFunctionType::LocalFunctionType LocalFunctionType;

  typedef typename DiscreteFunctionSpaceType::IteratorType IteratorType;
  typedef typename IteratorType::Entity EntityType;
  typedef typename EntityType::Geometry GeometryType;

  typedef typename DiscreteFunctionSpaceType::GridPartType GridPartType;
  typedef Dune::CachingQuadrature< GridPartType, 0 > QuadratureType;

public:
  explicit EllipticOperator ( const ModelType &model = Model() )
  : model_( model )
  {}
      
  virtual void
  operator() ( const DiscreteFunctionType &u, DiscreteFunctionType &w ) const;

  template< class JacobianOperator >
  void jacobian ( const DiscreteFunction &u, JacobianOperator &jOp ) const;

private:
  ModelType model_;
};

template< class DiscreteFunction, class Model >
inline void EllipticOperator< DiscreteFunction, Model >
  ::operator() ( const DiscreteFunctionType &u, DiscreteFunctionType &w ) const 
{
  w.clear();

  const DiscreteFunctionSpaceType &dfSpace = w.space();

  const IteratorType end = dfSpace.end();
  for( IteratorType it = dfSpace.begin(); it != end; ++it )
  {
    const EntityType &entity = *it;
    const GeometryType &geometry = entity.geometry();

    const LocalFunctionType uLocal = u.localFunction( entity );
    LocalFunctionType wLocal = w.localFunction( entity );

    QuadratureType quadrature( entity, uLocal.order() + wLocal.order() );
    const size_t numQuadraturePoints = quadrature.nop();
    for( size_t pt = 0; pt < numQuadraturePoints; ++pt )
    {
      const typename QuadratureType::CoordinateType &x = quadrature.point( pt );
      const double weight = quadrature.weight( pt ) * geometry.integrationElement( x );

      typename LocalFunctionType::RangeType vu, avu;
      uLocal.evaluate( quadrature[ pt ], vu );
      model_.massFlux( entity, quadrature[ pt ], vu, avu );

      typename LocalFunctionType::JacobianRangeType du, adu;
      uLocal.jacobian( quadrature[ pt ], du );
      model_.diffusiveFlux( entity, x, du, adu );

      avu *= weight;
      adu *= weight;
      wLocal.axpy( quadrature[ pt ], avu, adu );
    }
  }
  w.communicate();
}

template< class DiscreteFunction, class Model >
template< class JacobianOperator >
inline void EllipticOperator< DiscreteFunction, Model >
  ::jacobian ( const DiscreteFunction &u, JacobianOperator &jOp ) const
{
  typedef typename JacobianOperator::LocalMatrixType LocalMatrixType;
  typedef typename DiscreteFunctionSpaceType::BaseFunctionSetType BaseFunctionSetType;

  const DiscreteFunctionSpaceType &dfSpace = u.space();

  jOp.reserve();
  jOp.clear();

  std::vector< typename LocalFunctionType::RangeType > phi( dfSpace.mapper().maxNumDofs() );
  std::vector< typename LocalFunctionType::JacobianRangeType > dphi( dfSpace.mapper().maxNumDofs() );

  const IteratorType end = dfSpace.end();
  for( IteratorType it = dfSpace.begin(); it != end; ++it )
  {
    const EntityType &entity = *it;
    const GeometryType &geometry = entity.geometry();

    LocalMatrixType jLocal = jOp.localMatrix( entity, entity );

    const BaseFunctionSetType &baseSet = jLocal.domainBaseFunctionSet();
    const unsigned int numBaseFunctions = baseSet.size();
          
    QuadratureType quadrature( entity, 2*dfSpace.order() );
    const size_t numQuadraturePoints = quadrature.nop();
    for( size_t pt = 0; pt < numQuadraturePoints; ++pt )
    {
      const typename QuadratureType::CoordinateType &x = quadrature.point( pt );
      const double weight = quadrature.weight( pt ) * geometry.integrationElement( x );

      baseSet.evaluateAll( quadrature[ pt ], phi );
      const typename GeometryType::Jacobian &gjit = geometry.jacobianInverseTransposed( x );
      baseSet.jacobianAll( quadrature[ pt ], gjit, dphi );

      for( unsigned int i = 0; i < numBaseFunctions; ++i )
      {
        typename LocalFunctionType::RangeType aphi;
        model_.massFlux( entity, quadrature[ pt ], phi[ i ], aphi );
        typename LocalFunctionType::JacobianRangeType adphi;
        model_.diffusiveFlux( entity, x, dphi[ i ], adphi );
        for( unsigned int j = 0; j < numBaseFunctions; ++j )
        {
          double value = aphi * phi[ j ];
          for( int k = 0; k < adphi.rows; ++k )
            value += adphi[ k ] * dphi[ j ][ k ];
          jLocal.add( j, i, weight * value );
        }
      }
    }
  }
}

#endif // #ifndef ELLIPTIC_HH
