/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

#ifndef EQUI_HH
#define EQUI_HH

#include <dune/fem/quadrature/cachingquadrature.hh>
#include <dune/fem/io/parameter.hh>

// assembleRHS
// -----------

template< class DiscreteFunction >
void equilbrateSolution ( DiscreteFunction &rhs )
{
  DiscreteFunction equi( rhs );
  equi.clear();
  typedef typename DiscreteFunction::DiscreteFunctionSpaceType DiscreteFunctionSpaceType;
  typedef typename DiscreteFunction::LocalFunctionType LocalFunctionType;
  typedef typename DiscreteFunctionSpaceType :: RangeType RangeType;

  typedef typename DiscreteFunctionSpaceType::IteratorType IteratorType;
  typedef typename IteratorType::Entity EntityType;
  typedef typename EntityType::Geometry GeometryType;

  typedef typename DiscreteFunctionSpaceType::GridPartType GridPartType;
  typedef Dune::CachingQuadrature< GridPartType, 0 > QuadratureType;


  const DiscreteFunctionSpaceType &dfSpace = rhs.space();

  int nrOfVals =0;
  RangeType middel(0);
  
  const IteratorType end = dfSpace.end();
  for( IteratorType it = dfSpace.begin(); it != end; ++it )
  {
    const EntityType &entity = *it;

    LocalFunctionType rhsLocal = rhs.localFunction( entity );
    
    QuadratureType quadrature( entity, 2*dfSpace.order()+1 );
    const size_t numQuadraturePoints = quadrature.nop();
    for( size_t pt = 0; pt < numQuadraturePoints; ++pt )
    {
      RangeType val(0);
      rhsLocal.evaluate(quadrature[pt], val);

      if(val.two_norm()>1e-4)
      {
        middel += val;
        nrOfVals++;
      }
    }
  }

  middel /= (nrOfVals >0)? nrOfVals : 1;

  for( IteratorType it = dfSpace.begin(); it != end; ++it )
  {
    const EntityType &entity = *it;
    const GeometryType &geometry = entity.geometry();

    LocalFunctionType equiLocal = equi.localFunction( entity );
    
    QuadratureType quadrature( entity, 2*dfSpace.order()+1 );
    const size_t numQuadraturePoints = quadrature.nop();
    for( size_t pt = 0; pt < numQuadraturePoints; ++pt )
    {

      const typename QuadratureType::CoordinateType &x = quadrature.point( pt );
      double weight = quadrature.weight( pt ) * geometry.integrationElement( x );
      RangeType val(middel);
      val *= weight;

      equiLocal.axpy(quadrature[pt], val );
    }
  }
  const double factor = Dune::Parameter::getValue<double> ("windfield.equilbration.factor");
  equi *= factor;
  rhs += equi;
  rhs.communicate();
}

#endif // #ifndef RHS_HH
