/******************************************************************************

  dune-ash is a numerical simulation of the propagation of an ash cloud
  caused by a volcano eruption at a constant height in a given wind field.
  It was motivated by the eruptions of the Icelandic volcanoes in 2010 and
  2011, both leading to the closure of air space above several European
  countries.

  Copyright (C) 2012 - 2013 Janick Gerstenberger
  Copyright (C) 2011        Robert Kloefkorn
  Copyright (C) 2011 - 2013 Tobias Malkmus
  Copyright (C) 2011        David Nies
  Copyright (C) 2011 - 2013 Martin Nolte
  Copyright (C) 2011        Axel Pfeiffer
  Copyright (C) 2011 - 2012 Theresa Strauch

  dune-ash is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of
  the License, or (at your option) any later version.

  dune-ash is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this module. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

#ifndef BETWEENEND_INCLUDES
#define BETWEENEND_INCLUDES

#ifndef USE_TWISTFREE_MAPPER
#define USE_TWISTFREE_MAPPER
#endif

// iostream includes
#include <iostream>

// include grid part
#include <dune/fem/gridpart/adaptiveleafgridpart.hh>

// include discrete function space
#include <dune/fem/space/lagrangespace.hh>

// include discrete function
#include <dune/fem/function/adaptivefunction.hh>
#include <dune/fem/function/blockvectorfunction.hh>

// include linear operators
#include <dune/fem/operator/matrix/spmatrix.hh>
#include <dune/fem/operator/2order/lagrangematrixtraits.hh>

#include <dune/fem/solver/umfpacksolver.hh>
#include <dune/fem/solver/inverseoperators.hh>

// include Lagrange interpolation
#include <dune/fem/operator/lagrangeinterpolation.hh>

// include norms
#include <dune/fem/misc/l2norm.hh>

// include parameter handling
#include <dune/fem/io/parameter.hh>

#include <dune/fem/io/streams/xdrstreams.hh>

#include "../common/discretizationtraits.hh"
#include "windfield.hh"
#include "elliptic.hh"

namespace windfield
{
  typedef typename DiscretizationTraits::GridType GridType;
  typedef typename DiscretizationTraits::GridPartType GridPartType;

  typedef typename DiscretizationTraits::WindFieldSpaceType DiscreteFunctionSpaceType;
  typedef typename DiscreteFunctionSpaceType::FunctionSpaceType FunctionSpaceType;

  // choose linear operator
  typedef Dune::Fem::LagrangeMatrixTraits< DiscreteFunctionSpaceType > MatrixTraits;

  typedef typename DiscretizationTraits::WindFieldType DiscreteFunctionType;
  typedef Dune::SparseRowMatrixOperator< DiscreteFunctionType, DiscreteFunctionType, MatrixTraits > LinearOperatorType;
  typedef Dune::CGInverseOperator< DiscreteFunctionType > LinearInverseOperatorType;

  // select Lagrange interpolation
  typedef Dune::LagrangeInterpolation< DiscreteFunctionType > InterpolationType;
  
  typedef Problem::WindField< FunctionSpaceType > WindFieldType;
  typedef Problem::VortexFunction< FunctionSpaceType > VortexFunctionType;
  typedef Dune::GridFunctionAdapter< VortexFunctionType, GridPartType > VortexAdapterType;
  typedef Dune::GridFunctionAdapter< WindFieldType, GridPartType > WindFieldAdaptarType;

  // define Laplace operator
  typedef Problem::HeatModel< FunctionSpaceType > HeatModelType;
  typedef EllipticOperator< DiscreteFunctionType, HeatModelType > EllipticOperatorType;

  // select norm for error computation
  typedef Dune::L2Norm< GridPartType > NormType;

} // namespace windfield

#endif // #ifndef BETWEENEND_INCLUDES
